# Copyright 2014-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.tests.distance.test_distance_ratcliff_obershelp.

This module contains unit tests for abydos.distance.RatcliffObershelp
"""

import unittest
from difflib import SequenceMatcher

from abydos.distance import RatcliffObershelp

from .. import _corpus_file


class RatcliffObershelpTestCases(unittest.TestCase):
    """Test Ratcliff-Obserhelp functions.

    abydos.distance.RatcliffObershelp
    """

    cmp = RatcliffObershelp()

    def test_ratcliff_obershelp_sim(self):
        """Test abydos.distance.RatcliffObershelp.sim."""
        # https://github.com/rockymadden/stringmetric/blob/master/core/src/test/scala/com/rockymadden/stringmetric/similarity/RatcliffObershelpMetricSpec.scala
        self.assertEqual(self.cmp.sim('', ''), 1)
        self.assertEqual(self.cmp.sim('abc', ''), 0)
        self.assertEqual(self.cmp.sim('', 'xyz'), 0)
        self.assertEqual(self.cmp.sim('abc', 'abc'), 1)
        self.assertEqual(self.cmp.sim('123', '123'), 1)
        self.assertEqual(self.cmp.sim('abc', 'xyz'), 0)
        self.assertEqual(self.cmp.sim('123', '456'), 0)
        self.assertAlmostEqual(
            self.cmp.sim('aleksander', 'alexandre'), 0.7368421052631579
        )
        self.assertAlmostEqual(
            self.cmp.sim('alexandre', 'aleksander'), 0.7368421052631579
        )
        self.assertAlmostEqual(
            self.cmp.sim('pennsylvania', 'pencilvaneya'), 0.6666666666666666
        )
        self.assertAlmostEqual(
            self.cmp.sim('pencilvaneya', 'pennsylvania'), 0.6666666666666666
        )
        self.assertAlmostEqual(
            self.cmp.sim('abcefglmn', 'abefglmo'), 0.8235294117647058
        )
        self.assertAlmostEqual(
            self.cmp.sim('abefglmo', 'abcefglmn'), 0.8235294117647058
        )

        with open(_corpus_file('variantNames.csv')) as cav_testset:
            next(cav_testset)
            for line in cav_testset:
                line = line.strip().split(',')
                word1, word2 = line[0], line[4]
                self.assertAlmostEqual(
                    self.cmp.sim(word1, word2),
                    SequenceMatcher(None, word1, word2).ratio(),
                )

        with open(_corpus_file('wikipediaCommonMisspellings.csv')) as missp:
            next(missp)
            for line in missp:
                line = line.strip().upper()
                line = ''.join(
                    [
                        _
                        for _ in line.strip()
                        if _ in tuple('ABCDEFGHIJKLMNOPQRSTUVWXYZ,')
                    ]
                )
                word1, word2 = line.split(',')
                # print(word1, word2e)
                self.assertAlmostEqual(
                    self.cmp.sim(word1, word2),
                    SequenceMatcher(None, word1, word2).ratio(),
                )

    def test_ratcliff_obershelp_dist(self):
        """Test abydos.distance.RatcliffObershelp.dist."""
        # https://github.com/rockymadden/stringmetric/blob/master/core/src/test/scala/com/rockymadden/stringmetric/similarity/RatcliffObershelpMetricSpec.scala
        self.assertEqual(self.cmp.dist('', ''), 0)
        self.assertEqual(self.cmp.dist('abc', ''), 1)
        self.assertEqual(self.cmp.dist('', 'xyz'), 1)
        self.assertEqual(self.cmp.dist('abc', 'abc'), 0)
        self.assertEqual(self.cmp.dist('123', '123'), 0)
        self.assertEqual(self.cmp.dist('abc', 'xyz'), 1)
        self.assertEqual(self.cmp.dist('123', '456'), 1)
        self.assertAlmostEqual(
            self.cmp.dist('aleksander', 'alexandre'), 0.2631578947368421
        )
        self.assertAlmostEqual(
            self.cmp.dist('alexandre', 'aleksander'), 0.2631578947368421
        )
        self.assertAlmostEqual(
            self.cmp.dist('pennsylvania', 'pencilvaneya'), 0.3333333333333333
        )
        self.assertAlmostEqual(
            self.cmp.dist('pencilvaneya', 'pennsylvania'), 0.3333333333333333
        )
        self.assertAlmostEqual(
            self.cmp.dist('abcefglmn', 'abefglmo'), 0.1764705882352941
        )
        self.assertAlmostEqual(
            self.cmp.dist('abefglmo', 'abcefglmn'), 0.1764705882352941
        )


if __name__ == '__main__':
    unittest.main()



# Copyright 2014-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.tests.distance.test_distance_bag.

This module contains unit tests for abydos.distance.Bag
"""

import unittest

from abydos.distance import Bag
from abydos.tokenizer import SAPSTokenizer


class BagTestCases(unittest.TestCase):
    """Test bag similarity functions.

    abydos.distance.Bag
    """

    cmp = Bag()

    def test_bag_dist_abs(self):
        """Test abydos.distance.Bag.dist_abs."""
        self.assertEqual(self.cmp.dist_abs('', ''), 0)
        self.assertEqual(self.cmp.dist_abs('nelson', ''), 6)
        self.assertEqual(self.cmp.dist_abs('', 'neilsen'), 7)
        self.assertEqual(self.cmp.dist_abs('ab', 'a'), 1)
        self.assertEqual(self.cmp.dist_abs('ab', 'c'), 2)
        self.assertEqual(self.cmp.dist_abs('nelson', 'neilsen'), 2)
        self.assertEqual(self.cmp.dist_abs('neilsen', 'nelson'), 2)
        self.assertEqual(self.cmp.dist_abs('niall', 'neal'), 2)
        self.assertEqual(self.cmp.dist_abs('aluminum', 'Catalan'), 5)
        self.assertEqual(self.cmp.dist_abs('abcdefg', 'hijklm'), 7)
        self.assertEqual(self.cmp.dist_abs('abcdefg', 'hijklmno'), 8)

    def test_bag_sim(self):
        """Test abydos.distance.Bag.sim."""
        self.assertEqual(self.cmp.sim('', ''), 1)
        self.assertEqual(self.cmp.sim('nelson', ''), 0)
        self.assertEqual(self.cmp.sim('', 'neilsen'), 0)
        self.assertEqual(self.cmp.sim('ab', 'a'), 0.5)
        self.assertEqual(self.cmp.sim('ab', 'c'), 0)
        self.assertAlmostEqual(self.cmp.sim('nelson', 'neilsen'), 5 / 7)
        self.assertAlmostEqual(self.cmp.sim('neilsen', 'nelson'), 5 / 7)
        self.assertAlmostEqual(self.cmp.sim('niall', 'neal'), 3 / 5)
        self.assertAlmostEqual(self.cmp.sim('aluminum', 'Catalan'), 3 / 8)
        self.assertEqual(self.cmp.sim('abcdefg', 'hijklm'), 0)
        self.assertEqual(self.cmp.sim('abcdefg', 'hijklmno'), 0)

        self.assertEqual(Bag(tokenizer=SAPSTokenizer()).sim('DNA', 'RNA'), 0.5)

    def test_bag_dist(self):
        """Test abydos.distance.Bag.dist."""
        self.assertEqual(self.cmp.dist('', ''), 0)
        self.assertEqual(self.cmp.dist('nelson', ''), 1)
        self.assertEqual(self.cmp.dist('', 'neilsen'), 1)
        self.assertEqual(self.cmp.dist('ab', 'a'), 0.5)
        self.assertEqual(self.cmp.dist('ab', 'c'), 1)
        self.assertAlmostEqual(self.cmp.dist('nelson', 'neilsen'), 2 / 7)
        self.assertAlmostEqual(self.cmp.dist('neilsen', 'nelson'), 2 / 7)
        self.assertAlmostEqual(self.cmp.dist('niall', 'neal'), 2 / 5)
        self.assertAlmostEqual(self.cmp.dist('aluminum', 'Catalan'), 5 / 8)
        self.assertEqual(self.cmp.dist('abcdefg', 'hijklm'), 1)
        self.assertEqual(self.cmp.dist('abcdefg', 'hijklmno'), 1)


if __name__ == '__main__':
    unittest.main()



# Copyright 2019-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.tokenizer._c_or_v_cluster.

Consonant or vowel cluster tokenizer.

This tokenizer first performs wordpunct tokenization, so words are split into
separate units and non-letter characters are added as their own units.
Following this, words are further divided into strings of consonants only and
strings of vowels only.
"""

import re
import unicodedata

from typing import Callable, Optional, Set, Union

from ._tokenizer import _Tokenizer

__all__ = ['COrVClusterTokenizer']


class COrVClusterTokenizer(_Tokenizer):
    """A C- or V-cluster tokenizer.

    .. versionadded:: 0.4.0
    """

    def __init__(
        self,
        scaler: Optional[Union[str, Callable[[float], float]]] = None,
        consonants: Optional[Set[str]] = None,
        vowels: Optional[Set[str]] = None,
    ) -> None:
        """Initialize tokenizer.

        Parameters
        ----------
        scaler : None, str, or function
            A scaling function for the Counter:

                - None : no scaling
                - 'set' : All non-zero values are set to 1.
                - 'length' : Each token has weight equal to its length.
                - 'length-log' : Each token has weight equal to the log of its
                   length + 1.
                - 'length-exp' : Each token has weight equal to e raised to its
                   length.
                - a callable function : The function is applied to each value
                  in the Counter. Some useful functions include math.exp,
                  math.log1p, math.sqrt, and indexes into interesting integer
                  sequences such as the Fibonacci sequence.
        consonants : None or set(str)
            The set of characters to treat as consonants
        vowels : None or set(str)
            The set of characters to treat as vowels


        .. versionadded:: 0.4.0

        """
        super(COrVClusterTokenizer, self).__init__(scaler=scaler)
        if consonants:
            self._consonants = consonants
        else:
            self._consonants = set('bcdfghjklmnpqrstvwxzßBCDFGHJKLMNPQRSTVWXZ')
        if vowels:
            self._vowels = vowels
        else:
            self._vowels = set('aeiouyAEIOUY')
        self._regexp = re.compile(r'\w+|[^\w\s]+', flags=0)

    def tokenize(self, string: str) -> 'COrVClusterTokenizer':
        """Tokenize the term and store it.

        The tokenized term is stored as an ordered list and as a Counter
        object.

        Parameters
        ----------
        string : str
            The string to tokenize

        Examples
        --------
        >>> COrVClusterTokenizer().tokenize('seven-twelfths')
        COrVClusterTokenizer({'s': 1, 'e': 3, 'v': 1, 'n': 1, '-': 1, 'tw': 1,
        'lfths': 1})

        >>> COrVClusterTokenizer().tokenize('character')
        COrVClusterTokenizer({'ch': 1, 'a': 2, 'r': 2, 'ct': 1, 'e': 1})


        .. versionadded:: 0.4.0

        """
        self._string = string
        self._ordered_tokens = []
        token_list = self._regexp.findall(self._string)
        for token in token_list:
            if (
                token[0] not in self._consonants
                and token[0] not in self._vowels
            ):
                self._ordered_tokens.append(token)
            else:
                token = unicodedata.normalize('NFD', token)
                mode = 0  # 0 = starting mode, 1 = cons, 2 = vowels
                new_token = ''  # noqa: S105
                for char in token:
                    if char in self._consonants:
                        if mode == 2:
                            self._ordered_tokens.append(new_token)
                            new_token = char
                        else:
                            new_token += char
                        mode = 1
                    elif char in self._vowels:
                        if mode == 1:
                            self._ordered_tokens.append(new_token)
                            new_token = char
                        else:
                            new_token += char
                        mode = 2
                    else:  # This should cover combining marks, marks, etc.
                        new_token += char

                self._ordered_tokens.append(new_token)

        self._ordered_tokens = [
            unicodedata.normalize('NFC', token)
            for token in self._ordered_tokens
        ]
        self._scale_and_counterize()
        return self


if __name__ == '__main__':
    import doctest

    doctest.testmod(optionflags=doctest.NORMALIZE_WHITESPACE)



# Copyright 2018-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.tests.fuzz.test_phonetic.

This module contains fuzz tests for abydos.phonetic
"""

import codecs
import unittest
from random import choice, randint, sample

from abydos.phonetic import (
    Ainsworth,
    AlphaSIS,
    BeiderMorse,
    Caverphone,
    DaitchMokotoff,
    Davidson,
    Dolby,
    DoubleMetaphone,
    Eudex,
    FONEM,
    FuzzySoundex,
    Haase,
    HenryEarly,
    Koelner,
    LEIN,
    MRA,
    MetaSoundex,
    Metaphone,
    NRL,
    NYSIIS,
    Norphone,
    ONCA,
    PHONIC,
    PSHPSoundexFirst,
    PSHPSoundexLast,
    ParmarKumbharana,
    Phonem,
    Phonet,
    PhoneticSpanish,
    Phonex,
    Phonix,
    RefinedSoundex,
    RethSchek,
    RogerRoot,
    RussellIndex,
    SPFC,
    SfinxBis,
    SoundD,
    Soundex,
    SoundexBR,
    SpanishMetaphone,
    StatisticsCanada,
    Waahlin,
)

from . import EXTREME_TEST, _corpus_file, _fuzz, _random_char

spfc = SPFC()

algorithms = {
    'ainsworth': Ainsworth().encode,
    'alpha_sis': AlphaSIS().encode,
    'bmpm': BeiderMorse().encode,
    'bmpm_german': BeiderMorse(language_arg='german').encode,
    'bmpm_french': BeiderMorse(language_arg='french').encode,
    'bmpm_gen_exact': BeiderMorse(match_mode='exact').encode,
    'bmpm_ash_approx': BeiderMorse(name_mode='ash').encode,
    'bmpm_ash_exact': BeiderMorse(name_mode='ash', match_mode='exact').encode,
    'bmpm_sep_approx': BeiderMorse(name_mode='sep').encode,
    'bmpm_sep_exact': BeiderMorse(name_mode='sep', match_mode='exact').encode,
    'caverphone_1': Caverphone(version=1).encode,
    'caverphone_2': Caverphone().encode,
    'daitch_mokotoff_soundex': DaitchMokotoff().encode,
    'davidson': Davidson().encode,
    'dolby': Dolby().encode,
    'dolby_ml4': Dolby(max_length=4).encode,
    'dolby_vowels': Dolby(keep_vowels=True).encode,
    'double_metaphone': DoubleMetaphone().encode,
    'eudex': Eudex().encode,
    'fonem': FONEM().encode,
    'fuzzy_soundex': FuzzySoundex().encode,
    'fuzzy_soundex_0pad_ml8': FuzzySoundex(max_length=8, zero_pad=True).encode,
    'haase_phonetik': Haase().encode,
    'haase_phonetik_primary': Haase(primary_only=True).encode,
    'henry_early': HenryEarly().encode,
    'henry_early_ml8': HenryEarly(max_length=8).encode,
    'koelner_phonetik': Koelner().encode,
    'koelner_phonetik_alpha': Koelner().encode_alpha,
    'lein': LEIN().encode,
    'lein_nopad_ml8': LEIN(max_length=8, zero_pad=False).encode,
    'metasoundex': MetaSoundex().encode,
    'metasoundex_es': MetaSoundex(lang='es').encode,
    'metaphone': Metaphone().encode,
    'mra': MRA().encode,
    'norphone': Norphone().encode,
    'nrl': NRL().encode,
    'nysiis': NYSIIS().encode,
    'nysiis_modified': NYSIIS(modified=True).encode,
    'nysiis_ml_inf': NYSIIS(max_length=-1).encode,
    'onca': ONCA().encode,
    'onca_nopad_ml8': ONCA(max_length=8, zero_pad=False).encode,
    'parmar_kumbharana': ParmarKumbharana().encode,
    'phonem': Phonem().encode,
    'phonet_1': Phonet().encode,
    'phonet_2': Phonet(mode=2).encode,
    'phonet_1_none': Phonet(lang='none').encode,
    'phonet_2_none': Phonet(mode=2, lang='none').encode,
    'phonetic_spanish': PhoneticSpanish().encode,
    'phonetic_spanish_ml4': PhoneticSpanish(max_length=4).encode,
    'phonex': Phonex().encode,
    'phonex_0pad_ml6': Phonex(max_length=6, zero_pad=True).encode,
    'phonic': PHONIC().encode,
    'phonic_0pad_ml6': PHONIC(max_length=6, zero_pad=True).encode,
    'phonic_ext': PHONIC(extended=True).encode,
    'phonix': Phonix().encode,
    'phonix_0pad_ml6': Phonix(max_length=6, zero_pad=True).encode,
    'pshp_soundex_first': PSHPSoundexFirst().encode,
    'pshp_soundex_first_german': PSHPSoundexFirst(german=True).encode,
    'pshp_soundex_first_ml8': PSHPSoundexFirst(max_length=8).encode,
    'pshp_soundex_last': PSHPSoundexLast().encode,
    'pshp_soundex_last_german': PSHPSoundexLast(german=True).encode,
    'pshp_soundex_last_ml8': PSHPSoundexLast(max_length=8).encode,
    'refined_soundex': RefinedSoundex().encode,
    'refined_soundex_vowels': RefinedSoundex(retain_vowels=True).encode,
    'refined_soundex_0pad_ml6': RefinedSoundex(
        zero_pad=True, max_length=6
    ).encode,
    'reth_schek_phonetik': RethSchek().encode,
    'roger_root': RogerRoot().encode,
    'roger_root_nopad_ml8': RogerRoot(max_length=8, zero_pad=False).encode,
    'russell_index': RussellIndex().encode,
    'russell_index_alpha': RussellIndex().encode_alpha,
    'sfinxbis': SfinxBis().encode,
    'sfinxbis_ml6': SfinxBis(max_length=6).encode,
    'sound_d': SoundD().encode,
    'sound_d_ml8': SoundD(max_length=8).encode,
    'soundex': Soundex().encode,
    'soundex_reverse': Soundex(reverse=True).encode,
    'soundex_0pad_ml6': Soundex(zero_pad=True, max_length=6).encode,
    'soundex_special': Soundex(var='special').encode,
    'soundex_census': Soundex(var='Census').encode,
    'soundex_br': SoundexBR().encode,
    'spanish_metaphone': SpanishMetaphone().encode,
    'spanish_metaphone_modified': SpanishMetaphone(modified=True).encode,
    'spanish_metaphone_ml4': SpanishMetaphone(max_length=4).encode,
    'spfc': lambda _: spfc.encode('{0} {0}'.format(_)),
    'statistics_canada': StatisticsCanada().encode,
    'statistics_canada_ml8': StatisticsCanada(max_length=8).encode,
    'waahlin': Waahlin().encode,
    'waahlin_soundex': Waahlin(encoder=Soundex()).encode,
}


class BigListOfNaughtyStringsTestCases(unittest.TestCase):
    """Test each phonetic algorithm against the BLNS set.

    Here, we test each algorithm against each string, but we only care that it
    does not result in an exception.

    While not actually a fuzz test, this does serve the purpose of looking for
    errors resulting from unanticipated input.
    """

    def fuzz_test_blns(self):
        """Test each phonetic algorithm against the BLNS set."""
        blns = []
        omit_section = False
        with codecs.open(_corpus_file('blns.txt'), encoding='UTF-8') as nsf:
            for line in nsf:
                line = line[:-1]
                if 'Script Injection' in line:
                    omit_section = True
                if 'SQL Injection' in line:
                    omit_section = False
                if line and line[0] != '#':
                    bmpm_omit = omit_section | (len(line.split()) > 5)
                    blns.append((bmpm_omit, line))

        for algo in algorithms:
            for bmpm_omit, ns in blns:
                try:
                    if not (bmpm_omit and 'bmpm' in algo):
                        algorithms[algo](ns)
                except Exception as inst:
                    self.fail(
                        'Exception "{}" thrown by {} for BLNS: {}'.format(
                            inst, algo, ns
                        )
                    )


class FuzzedWordsTestCases(unittest.TestCase):
    """Test each phonetic algorithm against the base words set."""

    reps = 1000 * (10000 if EXTREME_TEST else 1)

    basewords = []
    with codecs.open(
        _corpus_file('basewords.txt'), encoding='UTF-8'
    ) as basewords_file:
        for line in basewords_file:
            line = line[:-1]
            if line:
                basewords.append(line)

    def fuzz_test_base(self):
        """Test each phonetic algorithm against the unfuzzed base words."""
        for algo in algorithms:
            for word in self.basewords:
                try:
                    if not ('bmpm' in algo and len(word) > 12):
                        algorithms[algo](word)
                except Exception as inst:
                    self.fail(
                        'Exception "{}" thrown by {} for word: {}'.format(
                            inst, algo, word
                        )
                    )

    def fuzz_test_20pct(self):
        """Fuzz test phonetic algorithms against 20% fuzzed words."""
        for _ in range(self.reps):
            fuzzed = _fuzz(choice(self.basewords), fuzziness=0.2)

            if EXTREME_TEST:
                algs = list(algorithms.keys())
            else:
                algs = sample(list(algorithms.keys()), k=5)

            for algo in algs:
                try:
                    if not ('bmpm' in algo and len(fuzzed) > 12):
                        algorithms[algo](fuzzed)
                except Exception as inst:
                    self.fail(
                        'Exception "{}" thrown by {} for word: {}'.format(
                            inst, algo, fuzzed
                        )
                    )

    def fuzz_test_100pct(self):
        """Fuzz test phonetic algorithms against 100% fuzzed words."""
        for _ in range(self.reps):
            fuzzed = _fuzz(choice(self.basewords), fuzziness=1)

            if EXTREME_TEST:
                algs = list(algorithms.keys())
            else:
                algs = sample(list(algorithms.keys()), k=5)

            for algo in algs:
                try:
                    if not ('bmpm' in algo and len(fuzzed) > 12):
                        algorithms[algo](fuzzed)
                except Exception as inst:
                    self.fail(
                        'Exception "{}" thrown by {} for word: {}'.format(
                            inst, algo, fuzzed
                        )
                    )

    def fuzz_test_fuzz_bmp(self):
        """Fuzz test phonetic algorithms against BMP fuzz."""
        for _ in range(self.reps):
            fuzzed = ''.join(
                _random_char(0xFFFF) for _ in range(0, randint(8, 16))
            )

            if EXTREME_TEST:
                algs = list(algorithms.keys())
            else:
                algs = sample(list(algorithms.keys()), k=5)

            for algo in algs:
                try:
                    algorithms[algo](fuzzed)
                except Exception as inst:
                    self.fail(
                        'Exception "{}" thrown by {} for word: {}'.format(
                            inst, algo, fuzzed
                        )
                    )

    def fuzz_test_fuzz_bmpsmp_letter(self):
        """Fuzz test phonetic algorithms against alphabetic BMP+SMP fuzz."""
        for _ in range(self.reps):
            fuzzed = ''.join(
                _random_char(0x1FFFF, ' LETTER ')
                for _ in range(0, randint(8, 16))
            )

            if EXTREME_TEST:
                algs = list(algorithms.keys())
            else:
                algs = sample(list(algorithms.keys()), k=5)

            for algo in algs:
                try:
                    algorithms[algo](fuzzed)
                except Exception as inst:
                    self.fail(
                        'Exception "{}" thrown by {} for word: {}'.format(
                            inst, algo, fuzzed
                        )
                    )

    def fuzz_test_fuzz_bmpsmp_latin(self):
        """Fuzz test phonetic algorithms against Latin BMP+SMP fuzz."""
        for _ in range(self.reps):
            fuzzed = ''.join(
                _random_char(0x1FFFF, 'LATIN ')
                for _ in range(0, randint(8, 16))
            )

            if EXTREME_TEST:
                algs = list(algorithms.keys())
            else:
                algs = sample(list(algorithms.keys()), k=5)

            for algo in algs:
                try:
                    algorithms[algo](fuzzed)
                except Exception as inst:
                    self.fail(
                        'Exception "{}" thrown by {} for word: {}'.format(
                            inst, algo, fuzzed
                        )
                    )

    def fuzz_test_fuzz_unicode(self):
        """Fuzz test phonetic algorithms against valid Unicode fuzz."""
        for _ in range(self.reps):
            fuzzed = ''.join(_random_char() for _ in range(0, randint(8, 16)))

            if EXTREME_TEST:
                algs = list(algorithms.keys())
            else:
                algs = sample(list(algorithms.keys()), k=5)

            for algo in algs:
                try:
                    algorithms[algo](fuzzed)
                except Exception as inst:
                    self.fail(
                        'Exception "{}" thrown by {} for word: {}'.format(
                            inst, algo, fuzzed
                        )
                    )


if __name__ == '__main__':
    unittest.main()



# Copyright 2019-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.tests.distance.test_distance_kuhns_ii.

This module contains unit tests for abydos.distance.KuhnsII
"""

import unittest

from abydos.distance import KuhnsII


class KuhnsIITestCases(unittest.TestCase):
    """Test KuhnsII functions.

    abydos.distance.KuhnsII
    """

    cmp = KuhnsII()
    cmp_no_d = KuhnsII(alphabet=0)

    def test_kuhns_ii_sim(self):
        """Test abydos.distance.KuhnsII.sim."""
        # Base cases
        self.assertEqual(self.cmp.sim('', ''), 0.3333333333333333)
        self.assertEqual(self.cmp.sim('a', ''), 0.3333333333333333)
        self.assertEqual(self.cmp.sim('', 'a'), 0.3333333333333333)
        self.assertEqual(self.cmp.sim('abc', ''), 0.3333333333333333)
        self.assertEqual(self.cmp.sim('', 'abc'), 0.3333333333333333)
        self.assertEqual(self.cmp.sim('abc', 'abc'), 0.9965986394557823)
        self.assertEqual(self.cmp.sim('abcd', 'efgh'), 0.32908163265306123)

        self.assertAlmostEqual(self.cmp.sim('Nigel', 'Niall'), 0.6615646259)
        self.assertAlmostEqual(self.cmp.sim('Niall', 'Nigel'), 0.6615646259)
        self.assertAlmostEqual(self.cmp.sim('Colin', 'Coiln'), 0.6615646259)
        self.assertAlmostEqual(self.cmp.sim('Coiln', 'Colin'), 0.6615646259)
        self.assertAlmostEqual(
            self.cmp.sim('ATCAACGAGT', 'AACGATTAG'), 0.7490723562
        )

        # Tests with alphabet=0 (no d factor)
        self.assertEqual(self.cmp_no_d.sim('', ''), 0.3333333333333333)
        self.assertEqual(self.cmp_no_d.sim('a', ''), 0.3333333333333333)
        self.assertEqual(self.cmp_no_d.sim('', 'a'), 0.3333333333333333)
        self.assertEqual(self.cmp_no_d.sim('abc', ''), 0.3333333333333333)
        self.assertEqual(self.cmp_no_d.sim('', 'abc'), 0.3333333333333333)
        self.assertEqual(self.cmp_no_d.sim('abc', 'abc'), 0.3333333333333333)
        self.assertEqual(self.cmp_no_d.sim('abcd', 'efgh'), 0.0)

        self.assertAlmostEqual(
            self.cmp_no_d.sim('Nigel', 'Niall'), 0.2222222222
        )
        self.assertAlmostEqual(
            self.cmp_no_d.sim('Niall', 'Nigel'), 0.2222222222
        )
        self.assertAlmostEqual(
            self.cmp_no_d.sim('Colin', 'Coiln'), 0.2222222222
        )
        self.assertAlmostEqual(
            self.cmp_no_d.sim('Coiln', 'Colin'), 0.2222222222
        )
        self.assertAlmostEqual(
            self.cmp_no_d.sim('ATCAACGAGT', 'AACGATTAG'), 0.2813852814
        )

    def test_kuhns_ii_dist(self):
        """Test abydos.distance.KuhnsII.dist."""
        # Base cases
        self.assertEqual(self.cmp.dist('', ''), 0.6666666666666667)
        self.assertEqual(self.cmp.dist('a', ''), 0.6666666666666667)
        self.assertEqual(self.cmp.dist('', 'a'), 0.6666666666666667)
        self.assertEqual(self.cmp.dist('abc', ''), 0.6666666666666667)
        self.assertEqual(self.cmp.dist('', 'abc'), 0.6666666666666667)
        self.assertEqual(self.cmp.dist('abc', 'abc'), 0.003401360544217691)
        self.assertEqual(self.cmp.dist('abcd', 'efgh'), 0.6709183673469388)

        self.assertAlmostEqual(self.cmp.dist('Nigel', 'Niall'), 0.3384353741)
        self.assertAlmostEqual(self.cmp.dist('Niall', 'Nigel'), 0.3384353741)
        self.assertAlmostEqual(self.cmp.dist('Colin', 'Coiln'), 0.3384353741)
        self.assertAlmostEqual(self.cmp.dist('Coiln', 'Colin'), 0.3384353741)
        self.assertAlmostEqual(
            self.cmp.dist('ATCAACGAGT', 'AACGATTAG'), 0.2509276438
        )

        # Tests with alphabet=0 (no d factor)
        self.assertEqual(self.cmp_no_d.dist('', ''), 0.6666666666666667)
        self.assertEqual(self.cmp_no_d.dist('a', ''), 0.6666666666666667)
        self.assertEqual(self.cmp_no_d.dist('', 'a'), 0.6666666666666667)
        self.assertEqual(self.cmp_no_d.dist('abc', ''), 0.6666666666666667)
        self.assertEqual(self.cmp_no_d.dist('', 'abc'), 0.6666666666666667)
        self.assertEqual(self.cmp_no_d.dist('abc', 'abc'), 0.6666666666666667)
        self.assertEqual(self.cmp_no_d.dist('abcd', 'efgh'), 1.0)

        self.assertAlmostEqual(
            self.cmp_no_d.dist('Nigel', 'Niall'), 0.7777777778
        )
        self.assertAlmostEqual(
            self.cmp_no_d.dist('Niall', 'Nigel'), 0.7777777778
        )
        self.assertAlmostEqual(
            self.cmp_no_d.dist('Colin', 'Coiln'), 0.7777777778
        )
        self.assertAlmostEqual(
            self.cmp_no_d.dist('Coiln', 'Colin'), 0.7777777778
        )
        self.assertAlmostEqual(
            self.cmp_no_d.dist('ATCAACGAGT', 'AACGATTAG'), 0.7186147186
        )

    def test_kuhns_ii_corr(self):
        """Test abydos.distance.KuhnsII.corr."""
        # Base cases
        self.assertEqual(self.cmp.corr('', ''), 0.0)
        self.assertEqual(self.cmp.corr('a', ''), 0.0)
        self.assertEqual(self.cmp.corr('', 'a'), 0.0)
        self.assertEqual(self.cmp.corr('abc', ''), 0.0)
        self.assertEqual(self.cmp.corr('', 'abc'), 0.0)
        self.assertEqual(self.cmp.corr('abc', 'abc'), 0.9948979591836735)
        self.assertEqual(self.cmp.corr('abcd', 'efgh'), -0.006377551020408163)

        self.assertAlmostEqual(self.cmp.corr('Nigel', 'Niall'), 0.4923469388)
        self.assertAlmostEqual(self.cmp.corr('Niall', 'Nigel'), 0.4923469388)
        self.assertAlmostEqual(self.cmp.corr('Colin', 'Coiln'), 0.4923469388)
        self.assertAlmostEqual(self.cmp.corr('Coiln', 'Colin'), 0.4923469388)
        self.assertAlmostEqual(
            self.cmp.corr('ATCAACGAGT', 'AACGATTAG'), 0.6236085343
        )

        # Tests with alphabet=0 (no d factor)
        self.assertEqual(self.cmp_no_d.corr('', ''), 0.0)
        self.assertEqual(self.cmp_no_d.corr('a', ''), 0.0)
        self.assertEqual(self.cmp_no_d.corr('', 'a'), 0.0)
        self.assertEqual(self.cmp_no_d.corr('abc', ''), 0.0)
        self.assertEqual(self.cmp_no_d.corr('', 'abc'), 0.0)
        self.assertEqual(self.cmp_no_d.corr('abc', 'abc'), 0.0)
        self.assertEqual(self.cmp_no_d.corr('abcd', 'efgh'), -0.5)

        self.assertAlmostEqual(
            self.cmp_no_d.corr('Nigel', 'Niall'), -0.1666666667
        )
        self.assertAlmostEqual(
            self.cmp_no_d.corr('Niall', 'Nigel'), -0.1666666667
        )
        self.assertAlmostEqual(
            self.cmp_no_d.corr('Colin', 'Coiln'), -0.1666666667
        )
        self.assertAlmostEqual(
            self.cmp_no_d.corr('Coiln', 'Colin'), -0.1666666667
        )
        self.assertAlmostEqual(
            self.cmp_no_d.corr('ATCAACGAGT', 'AACGATTAG'), -0.0779220779
        )


if __name__ == '__main__':
    unittest.main()



# Copyright 2018-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.tests.stemmer.test_stemmer_uealite.

This module contains unit tests for abydos.stemmer.UEALite
"""

import unittest

from abydos.stemmer import UEALite

from .. import _corpus_file


class UEALiteTestCases(unittest.TestCase):
    """Test UEA-lite functions.

    abydos.stemmer.UEALite
    """

    stmr = UEALite()
    stmr_adams = UEALite(var='Adams')

    def test_uealite(self):
        """Test abydos.stemmer.UEALite."""
        # base case
        self.assertEqual(self.stmr.stem(''), '')

        # test cases copied from Ruby port
        # https://github.com/ealdent/uea-stemmer/blob/master/test/uea_stemmer_test.rb
        # These are corrected to match the Java version's output.
        # stem base words to just the base word
        self.assertEqual(self.stmr.stem('man'), 'man')
        self.assertEqual(self.stmr.stem('happiness'), 'happiness')
        # stem theses as thesis but not bases as basis
        self.assertEqual(self.stmr.stem('theses'), 'thesis')
        self.assertNotEqual(self.stmr.stem('bases'), 'basis')
        # stem preterite words ending in -ed without the -ed
        self.assertEqual(self.stmr.stem('ordained'), 'ordain')
        self.assertEqual(self.stmr.stem('killed'), 'kill')
        self.assertEqual(self.stmr.stem('liked'), 'lik')
        self.assertEqual(self.stmr.stem('helped'), 'help')
        self.assertEqual(self.stmr.stem('scarred'), 'scarre')
        self.assertEqual(self.stmr.stem('invited'), 'invit')
        self.assertEqual(self.stmr.stem('exited'), 'exit')
        self.assertEqual(self.stmr.stem('debited'), 'debit')
        self.assertEqual(self.stmr.stem('smited'), 'smit')
        # stem progressive verbs and gerunds without the -ing
        self.assertEqual(self.stmr.stem('running'), 'run')
        self.assertEqual(self.stmr.stem('settings'), 'set')
        self.assertEqual(self.stmr.stem('timing'), 'time')
        self.assertEqual(self.stmr.stem('dying'), 'dy')
        self.assertEqual(self.stmr.stem('harping'), 'harp')
        self.assertEqual(self.stmr.stem('charring'), 'char')
        # not stem false progressive verbs such as 'sing'
        self.assertEqual(self.stmr.stem('ring'), 'ring')
        self.assertEqual(self.stmr.stem('sing'), 'se')
        self.assertEqual(self.stmr.stem('bring'), 'br')
        self.assertEqual(self.stmr.stem('fling'), 'fle')
        # stem various plural nouns and 3rd-pres verbs without the -s/-es
        self.assertEqual(self.stmr.stem('changes'), 'change')
        self.assertEqual(self.stmr.stem('deaths'), 'death')
        self.assertEqual(self.stmr.stem('shadows'), 'shadow')
        self.assertEqual(self.stmr.stem('flies'), 'fly')
        self.assertEqual(self.stmr.stem('things'), 'thing')
        self.assertEqual(self.stmr.stem('nothings'), 'nothing')
        self.assertEqual(self.stmr.stem('witches'), 'witch')
        self.assertEqual(self.stmr.stem('makes'), 'mak')
        self.assertEqual(self.stmr.stem('smokes'), 'smok')
        self.assertEqual(self.stmr.stem('does'), 'do')
        # stem various words with -des suffix
        self.assertEqual(self.stmr.stem('abodes'), 'abod')
        self.assertEqual(self.stmr.stem('escapades'), 'escapad')
        self.assertEqual(self.stmr.stem('crusades'), 'crusad')
        self.assertEqual(self.stmr.stem('grades'), 'grad')
        # stem various words with -res suffix
        self.assertEqual(self.stmr.stem('wires'), 'wir')
        self.assertEqual(self.stmr.stem('acres'), 'acr')
        self.assertEqual(self.stmr.stem('fires'), 'fir')
        self.assertEqual(self.stmr.stem('cares'), 'car')
        # stem acronyms when pluralized otherwise they should be left alone
        self.assertEqual(self.stmr.stem('USA'), 'USA')
        self.assertEqual(self.stmr.stem('FLOSS'), 'FLOSS')
        self.assertEqual(self.stmr.stem('MREs'), 'MRE')
        self.assertEqual(self.stmr.stem('USAED'), 'USAED')

        # test cases copied from Ruby port
        # https://github.com/ealdent/uea-stemmer/blob/master/test/uea_stemmer_test.rb
        # stem base words to just the base word
        self.assertEqual(self.stmr_adams.stem('man'), 'man')
        self.assertEqual(self.stmr_adams.stem('happiness'), 'happiness')
        # stem theses as thesis but not bases as basis
        self.assertEqual(self.stmr_adams.stem('theses'), 'thesis')
        self.assertNotEqual(self.stmr_adams.stem('bases'), 'basis')
        # stem preterite words ending in -ed without the -ed
        self.assertEqual(self.stmr_adams.stem('ordained'), 'ordain')
        self.assertEqual(self.stmr_adams.stem('killed'), 'kill')
        self.assertEqual(self.stmr_adams.stem('liked'), 'like')
        self.assertEqual(self.stmr_adams.stem('helped'), 'help')
        # self.assertEqual(self.stmr_adams.stem('scarred'), 'scar')
        self.assertEqual(self.stmr_adams.stem('invited'), 'invite')
        self.assertEqual(self.stmr_adams.stem('exited'), 'exit')
        self.assertEqual(self.stmr_adams.stem('debited'), 'debit')
        self.assertEqual(self.stmr_adams.stem('smited'), 'smite')
        # stem progressive verbs and gerunds without the -ing
        self.assertEqual(self.stmr_adams.stem('running'), 'run')
        self.assertEqual(self.stmr_adams.stem('settings'), 'set')
        self.assertEqual(self.stmr_adams.stem('timing'), 'time')
        self.assertEqual(self.stmr_adams.stem('dying'), 'die')
        self.assertEqual(self.stmr_adams.stem('harping'), 'harp')
        self.assertEqual(self.stmr_adams.stem('charring'), 'char')
        # not stem false progressive verbs such as 'sing'
        self.assertEqual(self.stmr_adams.stem('ring'), 'ring')
        self.assertEqual(self.stmr_adams.stem('sing'), 'sing')
        self.assertEqual(self.stmr_adams.stem('ring'), 'ring')
        self.assertEqual(self.stmr_adams.stem('bring'), 'bring')
        self.assertEqual(self.stmr_adams.stem('fling'), 'fling')
        # stem various plural nouns and 3rd-pres verbs without the -s/-es
        self.assertEqual(self.stmr_adams.stem('changes'), 'change')
        self.assertEqual(self.stmr_adams.stem('deaths'), 'death')
        self.assertEqual(self.stmr_adams.stem('shadows'), 'shadow')
        self.assertEqual(self.stmr_adams.stem('flies'), 'fly')
        self.assertEqual(self.stmr_adams.stem('things'), 'thing')
        self.assertEqual(self.stmr_adams.stem('nothings'), 'nothing')
        self.assertEqual(self.stmr_adams.stem('witches'), 'witch')
        self.assertEqual(self.stmr_adams.stem('makes'), 'make')
        self.assertEqual(self.stmr_adams.stem('smokes'), 'smoke')
        self.assertEqual(self.stmr_adams.stem('does'), 'do')
        # stem various words with -des suffix
        self.assertEqual(self.stmr_adams.stem('abodes'), 'abode')
        self.assertEqual(self.stmr_adams.stem('escapades'), 'escapade')
        self.assertEqual(self.stmr_adams.stem('crusades'), 'crusade')
        self.assertEqual(self.stmr_adams.stem('grades'), 'grade')
        # stem various words with -res suffix
        self.assertEqual(self.stmr_adams.stem('wires'), 'wire')
        self.assertEqual(self.stmr_adams.stem('acres'), 'acre')
        self.assertEqual(self.stmr_adams.stem('fires'), 'fire')
        self.assertEqual(self.stmr_adams.stem('cares'), 'care')
        # stem acronyms when pluralized otherwise they should be left alone
        self.assertEqual(self.stmr_adams.stem('USA'), 'USA')
        self.assertEqual(self.stmr_adams.stem('FLOSS'), 'FLOSS')
        self.assertEqual(self.stmr_adams.stem('MREs'), 'MRE')
        self.assertEqual(self.stmr_adams.stem('USAED'), 'USAED')

        # Perl version tests
        self.assertEqual(self.stmr.stem('ragings'), 'rage')
        self.assertEqual(UEALite(var='Perl').stem('ragings'), 'rag')

        # complete coverage
        self.assertEqual(self.stmr.stem('was'), 'was')
        self.assertEqual(self.stmr.stem('during'), 'during')
        self.assertEqual(
            UEALite(max_word_length=20).stem('abcdefghijklmnopqrstuvwxyz'),
            'abcdefghijklmnopqrstuvwxyz',
        )
        self.assertEqual(self.stmr.stem('10'), '10')
        self.assertEqual(self.stmr.stem('top-ten'), 'top-ten')
        self.assertEqual(self.stmr.stem('top-10'), 'top-10')
        self.assertEqual(self.stmr.stem('top_ten'), 'top_ten')
        self.assertEqual(
            UEALite(max_acro_length=8, var='Adams').stem('ABCDEFGHIJKLMs'),
            'ABCDEFGHIJKLMs',
        )
        self.assertEqual(
            UEALite(max_acro_length=8, var='Adams').stem('ABCDEFGHIJKLM'),
            'ABCDEFGHIJKLM',
        )
        self.assertEqual(self.stmr.stem('abcDefGhij'), 'abcDefGhij')
        self.assertEqual(self.stmr.stem('Tophat'), 'Tophat')

    def test_uealite_wsj_set(self):
        """Test abydos.stemmer.UEALite (WSJ testset)."""
        with open(_corpus_file('uea-lite_wsj.csv')) as wsj_ts:
            for wsj_line in wsj_ts:
                (word, uea, rule) = wsj_line.strip().split(',')
                self.assertEqual(
                    self.stmr._stem_and_rule(word),  # noqa: SF01
                    (uea, float(rule)),
                )


if __name__ == '__main__':
    unittest.main()



# Copyright 2019-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.tests.tokenizer.test_tokenizer_qgrams.

This module contains unit tests for abydos.tokenizer.QGrams
"""

import unittest

from abydos.tokenizer import CharacterTokenizer


class CharacterTokenizerTestCases(unittest.TestCase):
    """Test abydos.tokenizer.CharacterTokenizer."""

    def test_character_tokenizer(self):
        """Test abydos.tokenizer.CharacterTokenizer."""
        self.assertEqual(
            sorted(CharacterTokenizer().tokenize('').get_list()), []
        )
        self.assertEqual(
            sorted(CharacterTokenizer().tokenize('a').get_list()), ['a']
        )

        self.assertEqual(
            sorted(CharacterTokenizer().tokenize('NELSON').get_list()),
            sorted(['N', 'E', 'L', 'S', 'O', 'N']),
        )

    def test_character_tokenizer_intersections(self):
        """Test abydos.tokenizer.CharacterTokenizer intersections."""
        self.assertEqual(
            sorted(
                CharacterTokenizer().tokenize('NELSON')
                & CharacterTokenizer().tokenize('')
            ),
            [],
        )
        self.assertEqual(
            sorted(
                CharacterTokenizer().tokenize('')
                & CharacterTokenizer().tokenize('NEILSEN')
            ),
            [],
        )
        self.assertEqual(
            sorted(
                CharacterTokenizer().tokenize('NELSON')
                & CharacterTokenizer().tokenize('NEILSEN')
            ),
            sorted(['N', 'E', 'L', 'S']),
        )
        self.assertEqual(
            sorted(
                CharacterTokenizer().tokenize('NAIL')
                & CharacterTokenizer().tokenize('LIAN')
            ),
            sorted(['N', 'A', 'I', 'L']),
        )

    def test_character_tokenizer_counts(self):
        """Test abydos.tokenizer.CharacterTokenizer counts."""
        self.assertEqual(CharacterTokenizer().tokenize('').count(), 0)
        self.assertEqual(len(CharacterTokenizer().tokenize('').get_list()), 0)

        self.assertEqual(CharacterTokenizer().tokenize('NEILSEN').count(), 7)
        self.assertEqual(CharacterTokenizer().tokenize('NELSON').count(), 6)

        self.assertEqual(
            len(CharacterTokenizer().tokenize('NEILSEN').get_list()), 7
        )
        self.assertEqual(
            len(CharacterTokenizer().tokenize('NELSON').get_list()), 6
        )


if __name__ == '__main__':
    unittest.main()



# Copyright 2014-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.tests.distance.test_distance_gotoh.

This module contains unit tests for abydos.distance.Gotoh
"""

import unittest

from abydos.distance import Gotoh, NeedlemanWunsch

from .test_distance_needleman_wunsch import _sim_nw, _sim_wikipedia
from .. import NIALL


class GotohTestCases(unittest.TestCase):
    """Test Gotoh functions.

    abydos.distance.Gotoh
    """

    def test_gotoh_sim_score(self):
        """Test abydos.distance.Gotoh.sim_score."""
        self.assertEqual(Gotoh().sim_score('', ''), 0)

        # https://en.wikipedia.org/wiki/Needleman–Wunsch_algorithm
        self.assertEqual(
            Gotoh(1, 1, _sim_nw).sim_score('GATTACA', 'GCATGCU'), 0
        )
        self.assertGreaterEqual(
            Gotoh(1, 0.5, _sim_nw).sim_score('GATTACA', 'GCATGCU'),
            NeedlemanWunsch(1, _sim_nw).sim_score('GATTACA', 'GCATGCU'),
        )
        self.assertEqual(
            Gotoh(5, 5, _sim_wikipedia).sim_score('AGACTAGTTAC', 'CGAGACGT'),
            16,
        )
        self.assertGreaterEqual(
            Gotoh(5, 2, _sim_wikipedia).sim_score('AGACTAGTTAC', 'CGAGACGT'),
            NeedlemanWunsch(5, _sim_wikipedia).sim_score(
                'AGACTAGTTAC', 'CGAGACGT'
            ),
        )

        # checked against http://ds9a.nl/nwunsch/ (mismatch=1, gap=5, skew=5)
        self.assertEqual(
            Gotoh(5, 5, _sim_nw).sim_score('CGATATCAG', 'TGACGSTGC'), -5
        )
        self.assertGreaterEqual(
            Gotoh(5, 2, _sim_nw).sim_score('CGATATCAG', 'TGACGSTGC'),
            NeedlemanWunsch(5, _sim_nw).sim_score('CGATATCAG', 'TGACGSTGC'),
        )
        self.assertEqual(
            Gotoh(5, 5, _sim_nw).sim_score('AGACTAGTTAC', 'TGACGSTGC'), -7
        )
        self.assertGreaterEqual(
            Gotoh(5, 2, _sim_nw).sim_score('AGACTAGTTAC', 'TGACGSTGC'),
            NeedlemanWunsch(5, _sim_nw).sim_score('AGACTAGTTAC', 'TGACGSTGC'),
        )
        self.assertEqual(
            Gotoh(5, 5, _sim_nw).sim_score('AGACTAGTTAC', 'CGAGACGT'), -15
        )
        self.assertGreaterEqual(
            Gotoh(5, 2, _sim_nw).sim_score('AGACTAGTTAC', 'CGAGACGT'),
            NeedlemanWunsch(5, _sim_nw).sim_score('AGACTAGTTAC', 'CGAGACGT'),
        )

        # Test cases where one of the strings has length 1
        self.assertAlmostEqual(Gotoh().sim_score('', 'a'), -1.0)
        self.assertAlmostEqual(Gotoh().sim_score('abc', 'a'), -0.4)
        self.assertAlmostEqual(Gotoh().sim_score('a', ''), -1.0)
        self.assertAlmostEqual(Gotoh().sim_score('a', 'abc'), -0.4)

    def test_gotoh_sim_score_nialls(self):
        """Test abydos.distance.Gotoh.sim_score (Nialls set)."""
        # checked against http://ds9a.nl/nwunsch/ (mismatch=1, gap=2, skew=2)
        nw_vals = (5, 0, -2, 3, 1, 1, -2, -2, -1, -3, -3, -5, -3, -7, -7, -19)
        g22 = Gotoh(2, 2, _sim_nw)
        for i in range(len(NIALL)):
            self.assertEqual(g22.sim_score(NIALL[0], NIALL[i]), nw_vals[i])
        nw_vals2 = (5, 0, -2, 3, 1, 1, -2, -2, -1, -2, -3, -3, -2, -6, -6, -8)
        g21 = Gotoh(2, 1, _sim_nw)
        g205 = Gotoh(2, 0.5, _sim_nw)
        nw2 = NeedlemanWunsch(2, _sim_nw)
        for i in range(len(NIALL)):
            self.assertEqual(g21.sim_score(NIALL[0], NIALL[i]), nw_vals2[i])
            self.assertGreaterEqual(
                g205.sim_score(NIALL[0], NIALL[i]),
                nw2.sim_score(NIALL[0], NIALL[i]),
            )

    def test_gotoh_sim(self):
        """Test abydos.distance.Gotoh.sim."""
        self.assertEqual(Gotoh().sim('', ''), 1.0)

        # https://en.wikipedia.org/wiki/Needleman–Wunsch_algorithm
        self.assertEqual(Gotoh(1, 1, _sim_nw).sim('GATTACA', 'GCATGCU'), 0)
        self.assertGreaterEqual(
            Gotoh(1, 0.5, _sim_nw).sim('GATTACA', 'GCATGCU'),
            NeedlemanWunsch(1, _sim_nw).sim('GATTACA', 'GCATGCU'),
        )
        self.assertEqual(
            Gotoh(5, 5, _sim_wikipedia).sim('AGACTAGTTAC', 'CGAGACGT'),
            0.19950186722152657,
        )
        self.assertGreaterEqual(
            Gotoh(5, 2, _sim_wikipedia).sim('AGACTAGTTAC', 'CGAGACGT'),
            NeedlemanWunsch(5, _sim_wikipedia).sim('AGACTAGTTAC', 'CGAGACGT'),
        )


if __name__ == '__main__':
    unittest.main()



# Copyright 2014-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.tests.distance.test_distance_ident.

This module contains unit tests for abydos.distance.Ident
"""

import unittest

from abydos.distance import Ident


class IdentTestCases(unittest.TestCase):
    """Test identity similarity functions.

    abydos.distance.Ident
    """

    cmp = Ident()

    def test_ident_sim(self):
        """Test abydos.distance.Ident.sim."""
        self.assertEqual(self.cmp.sim('', ''), 1)
        self.assertEqual(self.cmp.sim('', 'a'), 0)
        self.assertEqual(self.cmp.sim('a', ''), 0)
        self.assertEqual(self.cmp.sim('a', 'a'), 1)
        self.assertEqual(self.cmp.sim('abcd', 'abcd'), 1)
        self.assertEqual(self.cmp.sim('abcd', 'dcba'), 0)
        self.assertEqual(self.cmp.sim('abc', 'cba'), 0)

    def test_ident_dist(self):
        """Test abydos.distance.Ident.dist."""
        self.assertEqual(self.cmp.dist('', ''), 0)
        self.assertEqual(self.cmp.dist('', 'a'), 1)
        self.assertEqual(self.cmp.dist('a', ''), 1)
        self.assertEqual(self.cmp.dist('a', 'a'), 0)
        self.assertEqual(self.cmp.dist('abcd', 'abcd'), 0)
        self.assertEqual(self.cmp.dist('abcd', 'dcba'), 1)
        self.assertEqual(self.cmp.dist('abc', 'cba'), 1)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# Copyright 2015-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""badge_update.py.

This updates the Pylint, pydocstyle, flake8, & SLOCCount badges in README.rst.
"""

import os.path
import re

BADGE_COLORS = (
    'brightgreen',
    'green',
    'yellowgreen',
    'yellow',
    'orange',
    'red',
)


def pylint_color(score):
    """Return Pylint badge color.

    Parameters
    ----------
    score : float
        A Pylint score

    Returns
    -------
    str
        Badge color

    """
    # These are the score cutoffs for each color above.
    # I.e. score==10 -> brightgreen, down to 7.5 > score >= 5 -> orange
    score_cutoffs = (10, 9.5, 8.5, 7.5, 5)
    for i in range(len(score_cutoffs)):
        if score >= score_cutoffs[i]:
            return BADGE_COLORS[i]
    # and score < 5 -> red
    return BADGE_COLORS[-1]


# def pycodestyle_color(score):
#     """Return pycodestyle badge color."""
#     # These are the score cutoffs for each color above.
#     # I.e. score==0 -> brightgreen, down to 100 < score <= 200 -> orange
#     score_cutoffs = (0, 20, 50, 100, 200)
#     for i in range(len(score_cutoffs)):
#         if score <= score_cutoffs[i]:
#             return BADGE_COLORS[i]
#     # and score > 200 -> red
#     return BADGE_COLORS[-1]


def pydocstyle_color(score):
    """Return pydocstyle badge color.

    Parameters
    ----------
    score : float
        A pydocstyle score

    Returns
    -------
    str
        Badge color

    """
    # These are the score cutoffs for each color above.
    # I.e. score==0 -> brightgreen, down to 100 < score <= 200 -> orange
    score_cutoffs = (0, 10, 25, 50, 100)
    for i in range(len(score_cutoffs)):
        if score <= score_cutoffs[i]:
            return BADGE_COLORS[i]
    # and score > 200 -> red
    return BADGE_COLORS[-1]


def flake8_color(score):
    """Return flake8 badge color.

    Parameters
    ----------
    score : float
        A flake8 score

    Returns
    -------
    str
        Badge color

    """
    # These are the score cutoffs for each color above.
    # I.e. score==0 -> brightgreen, down to 100 < score <= 200 -> orange
    score_cutoffs = (0, 20, 50, 100, 200)
    for i in range(len(score_cutoffs)):
        if score <= score_cutoffs[i]:
            return BADGE_COLORS[i]
    # and score > 200 -> red
    return BADGE_COLORS[-1]


if __name__ == '__main__':
    if not os.path.isfile('./README.rst'):
        exit('README.rst file missing.')

    if not os.path.isfile('./pylint.log'):
        exit('Please direct Pylint output to pylint.log')
    pylint_text = open('pylint.log', 'r', encoding='utf-8').read()
    pylint_score = max(
        float(
            re.search(
                'Your code has been rated at' + r' (-?[0-9\.]+)', pylint_text
            ).group(1)
        ),
        0.0,
    )

    # if not os.path.isfile('./pycodestyle.log'):
    #     exit('Please direct pycodestyle output to pycodestyle.log')
    # pycodestyle_text = open('pycodestyle.log', 'r', encoding='utf-8').read()
    # pycodestyle_score = sum(int(n) for n in re.findall(r'\n([0-9]+) +',
    #                                                    pycodestyle_text))

    if not os.path.isfile('./pydocstyle.log'):
        exit('Please direct pydocstyle output to pydocstyle.log')
    pydocstyle_score = int(
        open('pydocstyle.log', 'r', encoding='utf-8').read().split()[-1]
    )

    if not os.path.isfile('./flake8.log'):
        exit('Please direct flake8 output to flake8.log')
    flake8_score = int(
        open('flake8.log', 'r', encoding='utf-8').read().split()[-1]
    )

    if not os.path.isfile('./sloccount.log'):
        exit('Please direct sloccount output to sloccount.log')
    with open('sloccount.log', 'r', encoding='utf-8') as fh:
        pattern = r'Total Physical Source Lines of Code \(SLOC\) += ([0-9,]+)'
        sloccount = re.search(pattern, fh.read()).group(1)

    if not os.path.isfile('./mypy/index.txt'):
        exit('Please direct mypy txt report to mypy/index.txt')
    with open('mypy/index.txt', 'r', encoding='utf-8') as fh:
        pattern = r'\| Total +?\| +?([0-9.]+)'
        imprecision = re.search(pattern, fh.read()).group(1)

    readme_text = open('README.rst', 'r', encoding='utf-8').read()

    prefix = 'https://img.shields.io/badge/Pylint-'
    readme_text = re.sub(
        prefix + r'(-?[0-9\.]+/10-[a-z]+)',
        prefix + str(pylint_score) + '/10-' + pylint_color(pylint_score),
        readme_text,
        1,
    )

    # prefix = 'https://img.shields.io/badge/pycodestyle-'
    # readme_text = re.sub(prefix + r'([0-9\.]+-[a-z]+)',
    #                      prefix + str(pycodestyle_score) + '-' +
    #                      pycodestyle_color(pycodestyle_score),
    #                      readme_text, 1)

    prefix = 'https://img.shields.io/badge/pydocstyle-'
    readme_text = re.sub(
        prefix + r'([0-9\.]+-[a-z]+)',
        prefix
        + str(pydocstyle_score)
        + '-'
        + pydocstyle_color(pydocstyle_score),
        readme_text,
        1,
    )

    prefix = 'https://img.shields.io/badge/flake8-'
    readme_text = re.sub(
        prefix + r'([0-9\.]+-[a-z]+)',
        prefix + str(flake8_score) + '-' + flake8_color(flake8_score),
        readme_text,
        1,
    )

    prefix = 'https://img.shields.io/badge/SLOCCount-'
    readme_text = re.sub(
        prefix + r'[0-9,]+', prefix + str(sloccount), readme_text, 1
    )

    prefix = 'https://img.shields.io/badge/mypy-'
    readme_text = re.sub(
        prefix + r'[0-9.]+', prefix + str(imprecision), readme_text, 1
    )

    with open('README.rst', 'w', encoding='utf-8') as f:
        f.write(readme_text)



# Copyright 2014-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.tests.distance.test_distance_jaro_winkler.

This module contains unit tests for abydos.distance.JaroWinkler
"""

import unittest

from abydos.distance import JaroWinkler


class JaroWinklerTestCases(unittest.TestCase):
    """Test Jaro(-Winkler) functions.

    abydos.distance.JaroWinkler
    """

    jaro = JaroWinkler(mode='jaro')
    jaro_winkler = JaroWinkler(mode='winkler')

    def test_sim_jaro_winkler(self):
        """Test abydos.distance.JaroWinkler.sim."""
        self.assertEqual(self.jaro.sim('', ''), 1)
        self.assertEqual(self.jaro_winkler.sim('', ''), 1)
        self.assertEqual(self.jaro.sim('MARTHA', ''), 0)
        self.assertEqual(self.jaro_winkler.sim('MARTHA', ''), 0)
        self.assertEqual(self.jaro.sim('', 'MARHTA'), 0)
        self.assertEqual(self.jaro_winkler.sim('', 'MARHTA'), 0)
        self.assertEqual(self.jaro.sim('MARTHA', 'MARTHA'), 1)
        self.assertEqual(self.jaro_winkler.sim('MARTHA', 'MARTHA'), 1)

        # https://en.wikipedia.org/wiki/Jaro-Winkler_distance
        self.assertAlmostEqual(self.jaro.sim('MARTHA', 'MARHTA'), 0.94444444)
        self.assertAlmostEqual(
            self.jaro_winkler.sim('MARTHA', 'MARHTA'), 0.96111111
        )
        self.assertAlmostEqual(self.jaro.sim('DWAYNE', 'DUANE'), 0.82222222)
        self.assertAlmostEqual(self.jaro_winkler.sim('DWAYNE', 'DUANE'), 0.84)
        self.assertAlmostEqual(self.jaro.sim('DIXON', 'DICKSONX'), 0.76666666)
        self.assertAlmostEqual(
            self.jaro_winkler.sim('DIXON', 'DICKSONX'), 0.81333333
        )

        self.assertRaises(
            ValueError, JaroWinkler(boost_threshold=2).sim, 'abcd', 'dcba'
        )
        self.assertRaises(
            ValueError, JaroWinkler(boost_threshold=-1).sim, 'abcd', 'dcba'
        )
        self.assertRaises(
            ValueError, JaroWinkler(scaling_factor=0.3).sim, 'abcd', 'dcba'
        )
        self.assertRaises(
            ValueError, JaroWinkler(scaling_factor=-1).sim, 'abcd', 'dcba'
        )

        self.assertAlmostEqual(self.jaro_winkler.sim('ABCD', 'EFGH'), 0.0)

        # long_strings = True (applies only to Jaro-Winkler, not Jaro)
        self.assertEqual(
            JaroWinkler(long_strings=True).sim('ABCD', 'EFGH'),
            self.jaro.sim('ABCD', 'EFGH'),
        )
        self.assertEqual(
            JaroWinkler(mode='jaro', long_strings=True).sim(
                'DIXON', 'DICKSONX'
            ),
            self.jaro.sim('DIXON', 'DICKSONX'),
        )
        self.assertAlmostEqual(
            JaroWinkler(mode='winkler', long_strings=True).sim(
                'DIXON', 'DICKSONX'
            ),
            0.83030303,
        )
        self.assertAlmostEqual(
            JaroWinkler(mode='winkler', long_strings=True).sim(
                'MARTHA', 'MARHTA'
            ),
            0.97083333,
        )

    def test_dist_jaro_winkler(self):
        """Test abydos.distance.JaroWinkler.dist."""
        self.assertEqual(self.jaro.dist('', ''), 0)
        self.assertEqual(self.jaro_winkler.dist('', ''), 0)
        self.assertEqual(self.jaro.dist('MARTHA', ''), 1)
        self.assertEqual(self.jaro_winkler.dist('MARTHA', ''), 1)
        self.assertEqual(self.jaro.dist('', 'MARHTA'), 1)
        self.assertEqual(self.jaro_winkler.dist('', 'MARHTA'), 1)
        self.assertEqual(self.jaro.dist('MARTHA', 'MARTHA'), 0)
        self.assertEqual(self.jaro_winkler.dist('MARTHA', 'MARTHA'), 0)

        # https://en.wikipedia.org/wiki/Jaro-Winkler_distance
        self.assertAlmostEqual(self.jaro.dist('MARTHA', 'MARHTA'), 0.05555555)
        self.assertAlmostEqual(
            self.jaro_winkler.dist('MARTHA', 'MARHTA'), 0.03888888
        )
        self.assertAlmostEqual(self.jaro.dist('DWAYNE', 'DUANE'), 0.17777777)
        self.assertAlmostEqual(self.jaro_winkler.dist('DWAYNE', 'DUANE'), 0.16)
        self.assertAlmostEqual(self.jaro.dist('DIXON', 'DICKSONX'), 0.23333333)
        self.assertAlmostEqual(
            self.jaro_winkler.dist('DIXON', 'DICKSONX'), 0.18666666
        )

        self.assertRaises(
            ValueError, JaroWinkler(boost_threshold=2).dist, 'abcd', 'dcba'
        )
        self.assertRaises(
            ValueError, JaroWinkler(boost_threshold=-1).dist, 'abcd', 'dcba'
        )
        self.assertRaises(
            ValueError, JaroWinkler(scaling_factor=0.3).dist, 'abcd', 'dcba'
        )
        self.assertRaises(
            ValueError, JaroWinkler(scaling_factor=-1).dist, 'abcd', 'dcba'
        )

        self.assertAlmostEqual(self.jaro_winkler.dist('ABCD', 'EFGH'), 1.0)


if __name__ == '__main__':
    unittest.main()



# Copyright 2014-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.distance._ident.

Identity similarity & distance
"""

from ._distance import _Distance

__all__ = ['Ident']


class Ident(_Distance):
    """Identity distance and similarity.

    .. versionadded:: 0.3.6
    """

    def sim(self, src: str, tar: str) -> float:
        """Return the identity similarity of two strings.

        Identity similarity is 1.0 if the two strings are identical,
        otherwise 0.0

        Parameters
        ----------
        src : str
            Source string for comparison
        tar : str
            Target string for comparison

        Returns
        -------
        float
            Identity similarity

        Examples
        --------
        >>> cmp = Ident()
        >>> cmp.sim('cat', 'hat')
        0.0
        >>> cmp.sim('cat', 'cat')
        1.0


        .. versionadded:: 0.1.0
        .. versionchanged:: 0.3.6
            Encapsulated in class

        """
        return 1.0 if src == tar else 0.0


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# Copyright 2019-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.tests.fuzz.test_stemmer.

This module contains fuzz tests for abydos.stemmer
"""

import codecs
import unittest
from random import choice, randint, sample

from abydos.stemmer import (
    CLEFGerman,
    CLEFGermanPlus,
    CLEFSwedish,
    Caumanns,
    Lovins,
    PaiceHusk,
    Porter,
    Porter2,
    SStemmer,
    Schinke,
    SnowballDanish,
    SnowballDutch,
    SnowballGerman,
    SnowballNorwegian,
    SnowballSwedish,
    UEALite,
)

from . import EXTREME_TEST, _corpus_file, _fuzz, _random_char

algorithms = {
    'caumanns': Caumanns().stem,
    'clef_german': CLEFGerman().stem,
    'clef_german_plus': CLEFGermanPlus().stem,
    'clef_swedish': CLEFSwedish().stem,
    'lovins': Lovins().stem,
    'paice_husk': PaiceHusk().stem,
    'porter': Porter().stem,
    'porter_earlyenglish': Porter(early_english=True).stem,
    'porter2': Porter2().stem,
    'porter2_earlyenglish': Porter2(early_english=True).stem,
    's_stemmer': SStemmer().stem,
    'schinke': Schinke().stem,
    'snowball_danish': SnowballDanish().stem,
    'snowball_dutch': SnowballDutch().stem,
    'snowball_german': SnowballGerman().stem,
    'snowball_german_alt': SnowballGerman(alternate_vowels=True).stem,
    'snowball_norwegian': SnowballNorwegian().stem,
    'snowball_swedish': SnowballSwedish().stem,
    'uea_lite': UEALite().stem,
    'uea_lite_adams': UEALite(var='Adams').stem,
    'uea_lite_perl': UEALite(var='Perl').stem,
}


class BigListOfNaughtyStringsTestCases(unittest.TestCase):
    """Test each stemmer against the BLNS set.

    Here, we test each algorithm against each string, but we only care that it
    does not result in an exception.

    While not actually a fuzz test, this does serve the purpose of looking for
    errors resulting from unanticipated input.
    """

    def fuzz_test_blns(self):
        """Test each stemmer against the BLNS set."""
        blns = []
        with codecs.open(_corpus_file('blns.txt'), encoding='UTF-8') as nsf:
            for line in nsf:
                line = line[:-1]
                if line and line[0] != '#':
                    blns.append(line)

        for algo in algorithms:
            for ns in blns:
                try:
                    algorithms[algo](ns)
                except Exception as inst:
                    self.fail(
                        'Exception "{}" thrown by {} for BLNS: {}'.format(
                            inst, algo, ns
                        )
                    )


class FuzzedWordsTestCases(unittest.TestCase):
    """Test each stemmer against the base words set."""

    reps = 1000 * (10000 if EXTREME_TEST else 1)

    basewords = []
    with codecs.open(
        _corpus_file('basewords.txt'), encoding='UTF-8'
    ) as basewords_file:
        for line in basewords_file:
            line = line[:-1]
            if line:
                basewords.append(line)

    def fuzz_test_base(self):
        """Test each stemmer against the unfuzzed base words."""
        for algo in algorithms:
            for word in self.basewords:
                try:
                    algorithms[algo](word)
                except Exception as inst:
                    self.fail(
                        'Exception "{}" thrown by {} for word: {}'.format(
                            inst, algo, word
                        )
                    )

    def fuzz_test_20pct(self):
        """Fuzz test stemmers against 20% fuzzed words."""
        for _ in range(self.reps):
            fuzzed = _fuzz(choice(self.basewords), fuzziness=0.2)

            if EXTREME_TEST:
                algs = list(algorithms.keys())
            else:
                algs = sample(list(algorithms.keys()), k=5)

            for algo in algs:
                try:
                    algorithms[algo](fuzzed)
                except Exception as inst:
                    self.fail(
                        'Exception "{}" thrown by {} for word: {}'.format(
                            inst, algo, fuzzed
                        )
                    )

    def fuzz_test_100pct(self):
        """Fuzz test stemmers against 100% fuzzed words."""
        for _ in range(self.reps):
            fuzzed = _fuzz(choice(self.basewords), fuzziness=1)

            if EXTREME_TEST:
                algs = list(algorithms.keys())
            else:
                algs = sample(list(algorithms.keys()), k=5)

            for algo in algs:
                try:
                    algorithms[algo](fuzzed)
                except Exception as inst:
                    self.fail(
                        'Exception "{}" thrown by {} for word: {}'.format(
                            inst, algo, fuzzed
                        )
                    )

    def fuzz_test_fuzz_bmp(self):
        """Fuzz test stemmers against BMP fuzz."""
        for _ in range(self.reps):
            fuzzed = ''.join(
                _random_char(0xFFFF) for _ in range(0, randint(8, 16))
            )

            if EXTREME_TEST:
                algs = list(algorithms.keys())
            else:
                algs = sample(list(algorithms.keys()), k=5)

            for algo in algs:
                try:
                    algorithms[algo](fuzzed)
                except Exception as inst:
                    self.fail(
                        'Exception "{}" thrown by {} for word: {}'.format(
                            inst, algo, fuzzed
                        )
                    )

    def fuzz_test_fuzz_bmpsmp_letter(self):
        """Fuzz test stemmers against alphabetic BMP+SMP fuzz."""
        for _ in range(self.reps):
            fuzzed = ''.join(
                _random_char(0x1FFFF, ' LETTER ')
                for _ in range(0, randint(8, 16))
            )

            if EXTREME_TEST:
                algs = list(algorithms.keys())
            else:
                algs = sample(list(algorithms.keys()), k=5)

            for algo in algs:
                try:
                    algorithms[algo](fuzzed)
                except Exception as inst:
                    self.fail(
                        'Exception "{}" thrown by {} for word: {}'.format(
                            inst, algo, fuzzed
                        )
                    )

    def fuzz_test_fuzz_bmpsmp_latin(self):
        """Fuzz test stemmers against Latin BMP+SMP fuzz."""
        for _ in range(self.reps):
            fuzzed = ''.join(
                _random_char(0x1FFFF, 'LATIN ')
                for _ in range(0, randint(8, 16))
            )

            if EXTREME_TEST:
                algs = list(algorithms.keys())
            else:
                algs = sample(list(algorithms.keys()), k=5)

            for algo in algs:
                try:
                    algorithms[algo](fuzzed)
                except Exception as inst:
                    self.fail(
                        'Exception "{}" thrown by {} for word: {}'.format(
                            inst, algo, fuzzed
                        )
                    )

    def fuzz_test_fuzz_unicode(self):
        """Fuzz test stemmers against valid Unicode fuzz."""
        for _ in range(self.reps):
            fuzzed = ''.join(_random_char() for _ in range(0, randint(8, 16)))

            if EXTREME_TEST:
                algs = list(algorithms.keys())
            else:
                algs = sample(list(algorithms.keys()), k=5)

            for algo in algs:
                try:
                    algorithms[algo](fuzzed)
                except Exception as inst:
                    self.fail(
                        'Exception "{}" thrown by {} for word: {}'.format(
                            inst, algo, fuzzed
                        )
                    )


if __name__ == '__main__':
    unittest.main()



# Copyright 2014-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.tests.stemmer.test_stemmer_caumanns.

This module contains unit tests for abydos.stemmer.Caumanns
"""

import unittest

from abydos.stemmer import Caumanns


class CaumannsTestCases(unittest.TestCase):
    """Test Caumanns functions.

    abydos.stemmer.Caumanns
    """

    stmr = Caumanns()

    def test_caumanns(self):
        """Test abydos.stemmer.Caumanns."""
        # base case
        self.assertEqual(self.stmr.stem(''), '')

        # tests from Caumanns' description of the algorithm
        self.assertEqual(self.stmr.stem('singt'), 'sing')
        self.assertEqual(self.stmr.stem('singen'), 'sing')
        self.assertEqual(self.stmr.stem('beliebt'), 'belieb')
        self.assertEqual(self.stmr.stem('beliebtester'), 'belieb')
        self.assertEqual(self.stmr.stem('stören'), 'stor')
        self.assertEqual(self.stmr.stem('stöhnen'), 'stoh')
        self.assertEqual(self.stmr.stem('Kuß'), 'kuss')
        self.assertEqual(self.stmr.stem('Küsse'), 'kuss')
        self.assertEqual(self.stmr.stem('Verlierer'), 'verlier')
        self.assertEqual(self.stmr.stem('Verlies'), 'verlie')
        self.assertEqual(self.stmr.stem('Maus'), 'mau')
        self.assertEqual(self.stmr.stem('Mauer'), 'mau')
        self.assertEqual(self.stmr.stem('Störsender'), 'stor')

        # additional tests to achieve full coverage
        self.assertEqual(self.stmr.stem('Müllerinnen'), 'mullerin')
        self.assertEqual(self.stmr.stem('Matrix'), 'matrix')
        self.assertEqual(self.stmr.stem('Matrizen'), 'matrix')

    def test_caumanns_lucene(self):
        """Test abydos.stemmer.Caumanns (Lucene tests).

        Based on tests from
        https://svn.apache.org/repos/asf/lucene.net/trunk/test/contrib/Analyzers/De/data.txt
        This is presumably Apache-licensed.
        """
        # German special characters are replaced:
        self.assertEqual(self.stmr.stem('häufig'), 'haufig')
        self.assertEqual(self.stmr.stem('üor'), 'uor')
        self.assertEqual(self.stmr.stem('björk'), 'bjork')

        # here the stemmer works okay, it maps related words to the same stem:
        self.assertEqual(self.stmr.stem('abschließen'), 'abschliess')
        self.assertEqual(self.stmr.stem('abschließender'), 'abschliess')
        self.assertEqual(self.stmr.stem('abschließendes'), 'abschliess')
        self.assertEqual(self.stmr.stem('abschließenden'), 'abschliess')

        self.assertEqual(self.stmr.stem('Tisch'), 'tisch')
        self.assertEqual(self.stmr.stem('Tische'), 'tisch')
        self.assertEqual(self.stmr.stem('Tischen'), 'tisch')
        self.assertEqual(self.stmr.stem('geheimtür'), 'geheimtur')

        self.assertEqual(self.stmr.stem('Haus'), 'hau')
        self.assertEqual(self.stmr.stem('Hauses'), 'hau')
        self.assertEqual(self.stmr.stem('Häuser'), 'hau')
        self.assertEqual(self.stmr.stem('Häusern'), 'hau')
        # here's a case where overstemming occurs, i.e. a word is
        # mapped to the same stem as unrelated words:
        self.assertEqual(self.stmr.stem('hauen'), 'hau')

        # here's a case where understemming occurs, i.e. two related words
        # are not mapped to the same stem. This is the case with basically
        # all irregular forms:
        self.assertEqual(self.stmr.stem('Drama'), 'drama')
        self.assertEqual(self.stmr.stem('Dramen'), 'dram')

        # replace "ß" with 'ss':
        self.assertEqual(self.stmr.stem('Ausmaß'), 'ausmass')

        # fake words to test if suffixes are cut off:
        self.assertEqual(self.stmr.stem('xxxxxe'), 'xxxxx')
        self.assertEqual(self.stmr.stem('xxxxxs'), 'xxxxx')
        self.assertEqual(self.stmr.stem('xxxxxn'), 'xxxxx')
        self.assertEqual(self.stmr.stem('xxxxxt'), 'xxxxx')
        self.assertEqual(self.stmr.stem('xxxxxem'), 'xxxxx')
        self.assertEqual(self.stmr.stem('xxxxxer'), 'xxxxx')
        self.assertEqual(self.stmr.stem('xxxxxnd'), 'xxxxx')
        # the suffixes are also removed when combined:
        self.assertEqual(self.stmr.stem('xxxxxetende'), 'xxxxx')

        # words that are shorter than four charcters are not changed:
        self.assertEqual(self.stmr.stem('xxe'), 'xxe')
        # -em and -er are not removed from words shorter than five characters:
        self.assertEqual(self.stmr.stem('xxem'), 'xxem')
        self.assertEqual(self.stmr.stem('xxer'), 'xxer')
        # -nd is not removed from words shorter than six characters:
        self.assertEqual(self.stmr.stem('xxxnd'), 'xxxnd')


if __name__ == '__main__':
    unittest.main()



# Copyright 2018-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.


"""abydos.tests.fuzz.

This module contains fuzz tests for Abydos
"""

import os
import unicodedata
from random import choice, randint, random
from string import printable

from .. import EXTREME_TEST as SUPER_EXTREME_TEST
from .. import _corpus_file as _super_corpus_file

CORPORA = os.path.join(os.path.dirname(__file__), 'corpora')

EXTREME_TEST = SUPER_EXTREME_TEST  # inherit setting from base tests
EXTREME_TEST = False  # Set to True to test EVERY single case (NB: takes hours)


if not EXTREME_TEST and os.path.isfile(
    os.path.join(os.path.dirname(__file__), 'EXTREME_TEST')
):
    # EXTREME_TEST file detected -- switching to EXTREME_TEST mode...
    EXTREME_TEST = True
if not EXTREME_TEST and os.path.isfile(
    os.path.join(os.path.dirname(__file__), '..', 'EXTREME_TEST')
):
    # EXTREME_TEST file detected -- switching to EXTREME_TEST mode...
    EXTREME_TEST = True


def _corpus_file(name, corpora_dir=CORPORA):
    """Return the path to a corpus file.

    Parameters
    ----------
    name : str
        Corpus file
    corpora_dir : str
        The directory containing the corpora

    Returns
    -------
    str
        The path to the corpus file

    """
    return _super_corpus_file(name, corpora_dir)


def _random_char(below=0x10FFFF, must_be=None):
    """Generate a random Unicode character below U+{below}.

    Parameters
    ----------
    below : int
        Maximum Unicode value
    must_be : str
        A required part of the character name

    Returns
    -------
    str
        A character

    """
    while True:
        char = chr(randint(0, below))  # noqa: S311
        try:
            name = unicodedata.name(char)
            if must_be is None or must_be in name:
                return char
        except ValueError:
            pass


def _fuzz(word, fuzziness=0.2, must_be=None):
    """Fuzz a word with noise.

    Parameters
    ----------
    word : str
        A word to fuzz
    fuzziness : float
        How fuzzy to make the word
    must_be : str
        A required part of the character name

    Returns
    -------
    str
        A fuzzed word

    """
    while True:
        new_word = []
        for ch in word:
            if random() > fuzziness:  # noqa: S311
                new_word.append(ch)
            else:
                if random() > 0.5:  # noqa: S311
                    new_word.append(choice(printable))  # noqa: S311
                elif random() > 0.8:  # noqa: S311
                    new_word.append(
                        _random_char(0x10FFFF, must_be)
                    )  # noqa: S311
                else:
                    new_word.append(
                        _random_char(0xFFFF, must_be)
                    )  # noqa: S311
                if random() > 0.5:  # noqa: S311
                    new_word.append(ch)
        new_word = ''.join(new_word)
        if new_word != word:
            return new_word



# Copyright 2014-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.tests.distance.test_distance_hamming.

This module contains unit tests for abydos.distance.Hamming
"""

import unittest

from abydos.distance import Hamming


class HammingTestCases(unittest.TestCase):
    """Test Hamming functions.

    abydos.distance.Hamming
    """

    cmp = Hamming()
    cmp_no_diff = Hamming(False)

    def test_hamming_dist_abs(self):
        """Test abydos.distance.Hamming.dist_abs."""
        self.assertEqual(self.cmp.dist_abs('', ''), 0)
        self.assertEqual(self.cmp_no_diff.dist_abs('', ''), 0)

        self.assertEqual(self.cmp.dist_abs('a', ''), 1)
        self.assertEqual(self.cmp.dist_abs('a', 'a'), 0)
        self.assertEqual(self.cmp_no_diff.dist_abs('a', 'a'), 0)
        self.assertEqual(self.cmp.dist_abs('a', 'b'), 1)
        self.assertEqual(self.cmp_no_diff.dist_abs('a', 'b'), 1)
        self.assertEqual(self.cmp.dist_abs('abc', 'cba'), 2)
        self.assertEqual(self.cmp_no_diff.dist_abs('abc', 'cba'), 2)
        self.assertEqual(self.cmp.dist_abs('abc', ''), 3)
        self.assertEqual(self.cmp.dist_abs('bb', 'cbab'), 3)

        # test exception
        self.assertRaises(ValueError, self.cmp_no_diff.dist_abs, 'ab', 'a')

        # https://en.wikipedia.org/wiki/Hamming_distance
        self.assertEqual(self.cmp.dist_abs('karolin', 'kathrin'), 3)
        self.assertEqual(self.cmp.dist_abs('karolin', 'kerstin'), 3)
        self.assertEqual(self.cmp.dist_abs('1011101', '1001001'), 2)
        self.assertEqual(self.cmp.dist_abs('2173896', '2233796'), 3)

    def test_hamming_dist(self):
        """Test abydos.distance.Hamming.dist."""
        self.assertEqual(self.cmp.dist('', ''), 0)
        self.assertEqual(self.cmp_no_diff.dist('', ''), 0)

        self.assertEqual(self.cmp.dist('a', ''), 1)
        self.assertEqual(self.cmp.dist('a', 'a'), 0)
        self.assertEqual(self.cmp_no_diff.dist('a', 'a'), 0)
        self.assertEqual(self.cmp.dist('a', 'b'), 1)
        self.assertEqual(self.cmp_no_diff.dist('a', 'b'), 1)
        self.assertAlmostEqual(self.cmp.dist('abc', 'cba'), 2 / 3)
        self.assertAlmostEqual(self.cmp_no_diff.dist('abc', 'cba'), 2 / 3)
        self.assertEqual(self.cmp.dist('abc', ''), 1)
        self.assertAlmostEqual(self.cmp.dist('bb', 'cbab'), 3 / 4)

        # test exception
        self.assertRaises(ValueError, self.cmp_no_diff.dist, 'ab', 'a')

        # https://en.wikipedia.org/wiki/Hamming_distance
        self.assertAlmostEqual(self.cmp.dist('karolin', 'kathrin'), 3 / 7)
        self.assertAlmostEqual(self.cmp.dist('karolin', 'kerstin'), 3 / 7)
        self.assertAlmostEqual(self.cmp.dist('1011101', '1001001'), 2 / 7)
        self.assertAlmostEqual(self.cmp.dist('2173896', '2233796'), 3 / 7)

    def test_hamming_sim(self):
        """Test abydos.distance.Hamming.sim."""
        self.assertEqual(self.cmp.sim('', ''), 1)
        self.assertEqual(self.cmp_no_diff.sim('', ''), 1)

        self.assertEqual(self.cmp.sim('a', ''), 0)
        self.assertEqual(self.cmp.sim('a', 'a'), 1)
        self.assertEqual(self.cmp_no_diff.sim('a', 'a'), 1)
        self.assertEqual(self.cmp.sim('a', 'b'), 0)
        self.assertEqual(self.cmp_no_diff.sim('a', 'b'), 0)
        self.assertAlmostEqual(self.cmp.sim('abc', 'cba'), 1 / 3)
        self.assertAlmostEqual(self.cmp_no_diff.sim('abc', 'cba'), 1 / 3)
        self.assertEqual(self.cmp.sim('abc', ''), 0)
        self.assertAlmostEqual(self.cmp.sim('bb', 'cbab'), 1 / 4)

        # test exception
        self.assertRaises(ValueError, self.cmp_no_diff.sim, 'ab', 'a')

        # https://en.wikipedia.org/wiki/Hamming_distance
        self.assertAlmostEqual(self.cmp.sim('karolin', 'kathrin'), 4 / 7)
        self.assertAlmostEqual(self.cmp.sim('karolin', 'kerstin'), 4 / 7)
        self.assertAlmostEqual(self.cmp.sim('1011101', '1001001'), 5 / 7)
        self.assertAlmostEqual(self.cmp.sim('2173896', '2233796'), 4 / 7)


if __name__ == '__main__':
    unittest.main()



from typing import Any, Optional, Union, Tuple, List

from numpy import dtype

def maximum_sctype(t: dtype) -> dtype: ...
def issctype(rep: Any) -> bool: ...
def obj2sctype(rep: Any, default: Optional[Any] = ...) -> type: ...
def issubclass_(
    arg1: type, arg2: Union[type, Tuple[Union[type, Tuple], ...]]
) -> bool: ...
def issubsctype(
    arg1: type, arg2: Union[type, Tuple[Union[type, Tuple], ...]]
) -> bool: ...
def issubdtype(
    arg1: Union[object, type],
    arg2: Union[type, Tuple[Union[type, Tuple], ...]],
) -> bool: ...
def sctype2char(sctype: Any) -> str: ...
def find_common_type(
    array_types: Union[dtype, List[dtype]],
    scalar_types: Union[dtype, List[dtype]],
) -> dtype: ...



# Copyright 2018-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.tests.distance.test_distance_eudex.

This module contains unit tests for abydos.distance.Eudex
"""

import unittest

from abydos.distance import Eudex


def _yield_1():
    while True:
        yield 1


class EudexTestCases(unittest.TestCase):
    """Test Eudex distance functions.

    abydos.distance.Eudex
    """

    cmp = Eudex()

    def test_eudex_dist_abs(self):
        """Test abydos.distance.Eudex.dist_abs."""
        # Base cases
        self.assertEqual(self.cmp.dist_abs('', ''), 0)
        self.assertEqual(Eudex(None).dist_abs('', ''), 0)
        self.assertEqual(Eudex('fibonacci').dist_abs('', ''), 0)
        self.assertEqual(Eudex([10, 1, 1, 1]).dist_abs('', ''), 0)
        self.assertEqual(Eudex(_yield_1).dist_abs('', ''), 0)
        self.assertEqual(self.cmp.dist_abs('', '', normalized=True), 0)

        self.assertEqual(self.cmp.dist_abs('Niall', 'Niall'), 0)
        self.assertEqual(Eudex(None).dist_abs('Niall', 'Niall'), 0)
        self.assertEqual(Eudex('fibonacci').dist_abs('Niall', 'Niall'), 0)
        self.assertEqual(Eudex([10, 1, 1, 1]).dist_abs('Niall', 'Niall'), 0)
        self.assertEqual(Eudex(_yield_1).dist_abs('Niall', 'Niall'), 0)
        self.assertEqual(
            self.cmp.dist_abs('Niall', 'Niall', normalized=True), 0
        )

        self.assertEqual(self.cmp.dist_abs('Niall', 'Neil'), 2)
        self.assertEqual(Eudex(None).dist_abs('Niall', 'Neil'), 1)
        self.assertEqual(Eudex('fibonacci').dist_abs('Niall', 'Neil'), 2)
        self.assertEqual(Eudex([10, 1, 1, 1]).dist_abs('Niall', 'Neil'), 1)
        self.assertEqual(Eudex(_yield_1).dist_abs('Niall', 'Neil'), 1)
        self.assertAlmostEqual(
            self.cmp.dist_abs('Niall', 'Neil', normalized=True), 0.00098039
        )

        self.assertEqual(self.cmp.dist_abs('Niall', 'Colin'), 524)
        self.assertEqual(Eudex(None).dist_abs('Niall', 'Colin'), 10)
        self.assertEqual(Eudex('fibonacci').dist_abs('Niall', 'Colin'), 146)
        self.assertEqual(Eudex([10, 1, 1, 1]).dist_abs('Niall', 'Colin'), 42)
        self.assertEqual(Eudex(_yield_1).dist_abs('Niall', 'Colin'), 10)
        self.assertAlmostEqual(
            self.cmp.dist_abs('Niall', 'Colin', normalized=True), 0.25686274
        )

    def test_eudex_dist(self):
        """Test abydos.distance.Eudex.dist."""
        # Base cases
        self.assertEqual(self.cmp.dist('', ''), 0)
        self.assertEqual(Eudex(None).dist('', ''), 0)
        self.assertEqual(Eudex('fibonacci').dist('', ''), 0)

        self.assertEqual(self.cmp.dist('Niall', 'Niall'), 0)
        self.assertEqual(Eudex(None).dist('Niall', 'Niall'), 0)
        self.assertEqual(Eudex('fibonacci').dist('Niall', 'Niall'), 0)

        self.assertAlmostEqual(self.cmp.dist('Niall', 'Neil'), 0.00098039)
        self.assertAlmostEqual(Eudex(None).dist('Niall', 'Neil'), 0.11111111)
        self.assertAlmostEqual(
            Eudex('fibonacci').dist('Niall', 'Neil'), 0.00287356
        )

        self.assertAlmostEqual(self.cmp.dist('Niall', 'Colin'), 0.25686275)
        self.assertAlmostEqual(Eudex(None).dist('Niall', 'Colin'), 0.16666667)
        self.assertAlmostEqual(
            Eudex('fibonacci').dist('Niall', 'Colin'), 0.20977011
        )

        with self.assertRaises(ValueError):
            Eudex('veryLarge').dist_abs('Niall', 'Colin')

    def test_eudex_sim(self):
        """Test abydos.distance.Eudex.sim."""
        # Base cases
        self.assertEqual(self.cmp.sim('', ''), 1)
        self.assertEqual(Eudex(None).sim('', ''), 1)
        self.assertEqual(Eudex('fibonacci').sim('', ''), 1)

        self.assertEqual(self.cmp.sim('Niall', 'Niall'), 1)
        self.assertEqual(Eudex(None).sim('Niall', 'Niall'), 1)
        self.assertEqual(Eudex('fibonacci').sim('Niall', 'Niall'), 1)

        self.assertAlmostEqual(self.cmp.sim('Niall', 'Neil'), 0.99901961)
        self.assertAlmostEqual(Eudex(None).sim('Niall', 'Neil'), 0.88888889)
        self.assertAlmostEqual(
            Eudex('fibonacci').sim('Niall', 'Neil'), 0.99712644
        )

        self.assertAlmostEqual(self.cmp.sim('Niall', 'Colin'), 0.74313725)
        self.assertAlmostEqual(Eudex(None).sim('Niall', 'Colin'), 0.83333333)
        self.assertAlmostEqual(
            Eudex('fibonacci').sim('Niall', 'Colin'), 0.79022989
        )


if __name__ == '__main__':
    unittest.main()



# Copyright 2014-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.tests.distance.test_distance_strcmp95.

This module contains unit tests for abydos.distance.Strcmp95
"""

import unittest

from abydos.distance import Strcmp95


class Strcmp95TestCases(unittest.TestCase):
    """Test Strcmp95 functions.

    abydos.distance.Strcmp95
    """

    cmp = Strcmp95()
    cmp_ls = Strcmp95(True)

    def test_strcmp95_sim(self):
        """Test abydos.distance.Strcmp95.sim."""
        self.assertEqual(self.cmp.sim('', ''), 1)
        self.assertEqual(self.cmp.sim('MARTHA', ''), 0)
        self.assertEqual(self.cmp.sim('', 'MARTHA'), 0)
        self.assertEqual(self.cmp.sim('MARTHA', 'MARTHA'), 1)

        self.assertAlmostEqual(self.cmp.sim('MARTHA', 'MARHTA'), 0.96111111)
        self.assertAlmostEqual(self.cmp.sim('DWAYNE', 'DUANE'), 0.873)
        self.assertAlmostEqual(self.cmp.sim('DIXON', 'DICKSONX'), 0.839333333)

        self.assertAlmostEqual(self.cmp.sim('ABCD', 'EFGH'), 0.0)

        # long_strings = True
        self.assertAlmostEqual(
            self.cmp_ls.sim('DIXON', 'DICKSONX'), 0.85393939
        )
        self.assertAlmostEqual(self.cmp_ls.sim('DWAYNE', 'DUANE'), 0.89609090)
        self.assertAlmostEqual(self.cmp_ls.sim('MARTHA', 'MARHTA'), 0.97083333)

        # cover case where we don't boost, etc.
        self.assertAlmostEqual(self.cmp.sim('A', 'ABCDEFGHIJK'), 69 / 99)
        self.assertAlmostEqual(self.cmp_ls.sim('A', 'ABCDEFGHIJK'), 69 / 99)
        self.assertAlmostEqual(self.cmp.sim('d', 'abcdefgh'), 0.708333333)
        self.assertAlmostEqual(self.cmp_ls.sim('d', 'abcdefgh'), 0.708333333)
        self.assertAlmostEqual(self.cmp_ls.sim('1', 'abc1efgh'), 0.708333333)
        self.assertAlmostEqual(
            self.cmp_ls.sim('12hundredths', '12hundred'), 0.916666667
        )

    def test_strcmp95_dist(self):
        """Test abydos.distance.Strcmp95.dist."""
        self.assertEqual(self.cmp.dist('', ''), 0)
        self.assertEqual(self.cmp.dist('MARTHA', ''), 1)
        self.assertEqual(self.cmp.dist('', 'MARTHA'), 1)
        self.assertEqual(self.cmp.dist('MARTHA', 'MARTHA'), 0)

        self.assertAlmostEqual(self.cmp.dist('MARTHA', 'MARHTA'), 0.03888888)
        self.assertAlmostEqual(self.cmp.dist('DWAYNE', 'DUANE'), 0.127)
        self.assertAlmostEqual(self.cmp.dist('DIXON', 'DICKSONX'), 0.160666666)

        self.assertAlmostEqual(self.cmp.dist('ABCD', 'EFGH'), 1.0)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# Copyright 2014-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""features_csv_to_dict.py.

This script converts a CSV document of feature values to a Python dict.

The CSV document is of the format
<phonetic symbol(s)>,<variant number>,<segmental>,<feature>+

    Phonetic symbols are IPA (or other system) symbols.
    Variant number refers to one of the following codes:
         0 = IPA
         1 = Americanist
         2 = IPA variant
         3 = Americanist variant
         9 = other
    Segmental must be either 1 (segmental) or 0 (featural). Featural symbols
    replace features on the preceding segmental symbol.
    Features may be 1 (+), -1 (-), or 0 (0) to indicate feature values.

Lines beginning with # are interpreted as comments
"""

import codecs
import getopt
import sys
import unicodedata


def main(argv):
    """Read input file and write to output.

    Parameters
    ----------
    argv : list
        Arguments to the script

    """
    first_col = 3
    last_col = -1

    def print_usage():
        """Print usage statement."""
        sys.stdout.write(
            'features_csv_to_dict.py -i <inputfile> ' + '[-o <outputfile>]\n'
        )
        sys.exit(2)

    def binarize(num):
        """Replace 0, -1, 1, 2 with 00, 10, 01, 11.

        Parameters
        ----------
        num : str
            The number to binarize

        Returns
        -------
        str
            A binarized number

        """
        if num == '0':  # 0
            return '00'
        elif num == '-1':  # -
            return '10'
        elif num == '1':  # +
            return '01'
        # '2' -> ± (segmental) or copy from base (non-segmental)
        return '11'

    def init_termdicts():
        """Initialize the terms dict.

        Returns
        -------
        (dict, dict)
            Term & feature mask dictionaries

        """
        ifile = codecs.open('features_terms.csv', 'r', 'utf-8')

        feature_mask = {}
        keyline = ifile.readline().strip().split(',')[first_col:last_col]
        mag = len(keyline)
        for i in range(len(keyline)):
            features = '0b' + ('00' * i) + '11' + ('00' * (mag - i - 1))
            feature_mask[keyline[i]] = int(features, 2)

        termdict = {}
        for line in ifile:
            line = line.strip().rstrip(',')
            if '#' in line:
                line = line[: line.find('#')].strip()
            if line:
                line = line.split(',')
                term = line[last_col]
                features = '0b' + ''.join(
                    [binarize(val) for val in line[first_col:last_col]]
                )
                termdict[term] = int(features, 2)

        return termdict, feature_mask

    def check_terms(sym, features, name, termdict):
        """Check terms.

        Check each term of the phone name to confirm that it matches
        the expected features implied by that feature.

        Parameters
        ----------
        sym : str
            Symbol to check
        features : int
            Phone features
        name : str
            Phone name
        termdict : dict
            Dictionary of terms

        """
        if '#' in name:
            name = name[: name.find('#')].strip()
        for term in name.split():
            if term in termdict:
                if termdict[term] & features != termdict[term]:
                    sys.stdout.write(
                        'Feature mismatch for term "'
                        + term
                        + '" in   '
                        + sym
                        + '\n'
                    )
            else:
                sys.stdout.write(
                    'Unknown term "'
                    + term
                    + '" in '
                    + name
                    + ' : '
                    + sym
                    + '\n'
                )

    def check_entailments(sym, features, feature_mask):
        """Check entailments.

        Check for necessary feature assignments (entailments)
        For example, [+round] necessitates [+labial].

        Parameters
        ----------
        sym : str
            Symbol to check
        features : int
            Phone features
        feature_mask : dict
            The feature mask

        """
        entailments = {
            '+labial': ('±round', '±protruded', '±compressed', '±labiodental'),
            '-labial': ('0round', '0protruded', '0compressed', '0labiodental'),
            '+coronal': ('±anterior', '±distributed'),
            '-coronal': ('0anterior', '0distributed'),
            '+dorsal': ('±high', '±low', '±front', '±back', '±tense'),
            '-dorsal': ('0high', '0low', '0front', '0back', '0tense'),
            '+pharyngeal': ('±atr', '±rtr'),
            '-pharyngeal': ('0atr', '0rtr'),
            '+protruded': ('+labial', '+round', '-compressed'),
            '+compressed': ('+labial', '+round', '-protruded'),
            '+glottalic_suction': ('-velaric_suction',),
            '+velaric_suction': ('-glottalic_suction',),
        }

        for feature in entailments:
            fname = feature[1:]
            if feature[0] == '+':
                fm = (feature_mask[fname] >> 1) & feature_mask[fname]
            else:
                fm = (feature_mask[fname] << 1) & feature_mask[fname]
            if (features & fm) == fm:
                for ent in entailments[feature]:
                    ename = ent[1:]
                    if ent[0] == '+':
                        efm = (feature_mask[ename] >> 1) & feature_mask[ename]
                    elif ent[0] == '-':
                        efm = (feature_mask[ename] << 1) & feature_mask[ename]
                    elif ent[0] == '0':
                        efm = 0
                    elif ent[0] == '±':
                        efm = feature_mask[ename]

                    if ent[0] == '±':
                        if (features & efm) == 0:
                            sys.stdout.write(
                                'Incorrect entailment for '
                                + sym
                                + ' for feature '
                                + fname
                                + ' and entailment '
                                + ename
                            )
                    else:
                        if (features & efm) != efm:
                            sys.stdout.write(
                                'Incorrect entailment for '
                                + sym
                                + ' for feature '
                                + fname
                                + ' and entailment '
                                + ename
                            )

    checkdict = {}  # a mapping of symbol to feature
    checkset_s = set()  # a set of the symbols seen
    checkset_f = set()  # a set of the feature values seen

    termdict, feature_mask = init_termdicts()

    ifile = ''
    ofile = ''
    try:
        opts = getopt.getopt(argv, 'hi:o:', ['ifile=', 'ofile='])[0]
    except getopt.GetoptError:
        print_usage()
    for opt, arg in opts:
        if opt == '-h':
            print_usage()
        elif opt in ('-i', '--ifile'):
            ifile = codecs.open(arg, 'r', 'utf-8')
        elif opt in ('-o', '--ofile'):
            ofile = codecs.open(arg, 'w', 'utf-8')
    if not ifile:
        print_usage()

    oline = 'PHONETIC_FEATURES = {'
    if not ofile:
        ofile = sys.stdout

    ofile.write(oline + '\n')

    keyline = ifile.readline().strip().split(',')[first_col:last_col]
    for line in ifile:
        line = line.strip().rstrip(',')

        if line.startswith('####'):
            break

        line = unicodedata.normalize('NFC', line)

        if not line or line.startswith('#'):
            oline = '                     ' + line

        else:
            line = line.strip().split(',')
            if '#' in line:
                line = line[: line.find('#')]
            symbol = line[0]
            variant = int(line[1])
            segmental = bool(line[2])
            features = '0b' + ''.join(
                [binarize(val) for val in line[first_col:last_col]]
            )
            name = line[-1].strip()
            if not segmental:
                features = '-' + features

            featint = int(features, 2)
            check_terms(symbol, featint, name, termdict)
            check_entailments(symbol, featint, feature_mask)
            if symbol in checkset_s:
                sys.stdout.write(
                    'Symbol ' + symbol + ' appears twice in CSV.\n'
                )
            else:
                checkset_s.add(symbol)

            if variant < 2:
                if featint in checkset_f:
                    sys.stdout.write(
                        'Feature set '
                        + str(featint)
                        + ' appears in CSV for two primary IPA '
                        + 'symbols: '
                        + symbol
                        + ' and '
                        + checkdict[featint]
                    )
                else:
                    checkdict[featint] = symbol
                    checkset_f.add(featint)

            if variant < 5:
                oline = "                     '{}': {},".format(
                    symbol, featint
                )
            else:
                oline = ''

        if oline:
            ofile.write(oline + '\n')

    ofile.write('                    }\n\nFEATURE_MASK = {')

    mag = len(keyline)
    for i in range(len(keyline)):
        features = int('0b' + ('00' * i) + '11' + ('00' * (mag - i - 1)), 2)
        oline = "                '{}': {},".format(keyline[i], features)
        ofile.write(oline + '\n')

    ofile.write('               }\n')


if __name__ == '__main__':
    main(sys.argv[1:])



# Copyright 2014-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.distance._smith_waterman.

Smith-Waterman score
"""

from typing import Any, Callable, Optional, cast

from numpy import float_ as np_float
from numpy import zeros as np_zeros

from ._needleman_wunsch import NeedlemanWunsch

__all__ = ['SmithWaterman']


class SmithWaterman(NeedlemanWunsch):
    """Smith-Waterman score.

    The Smith-Waterman score :cite:`Smith:1981` is a standard edit distance
    measure, differing from Needleman-Wunsch in that it focuses on local
    alignment and disallows negative scores.

    .. versionadded:: 0.3.6
    """

    def __init__(
        self,
        gap_cost: float = 1.0,
        sim_func: Optional[Callable[[str, str], float]] = None,
        **kwargs: Any
    ) -> None:
        """Initialize SmithWaterman instance.

        Parameters
        ----------
        gap_cost : float
            The cost of an alignment gap (1 by default)
        sim_func : function
            A function that returns the similarity of two characters (identity
            similarity by default)
        **kwargs
            Arbitrary keyword arguments


        .. versionadded:: 0.4.0

        """
        super(SmithWaterman, self).__init__(**kwargs)
        self._gap_cost = gap_cost
        self._sim_func = cast(
            Callable[[str, str], float],
            NeedlemanWunsch.sim_matrix if sim_func is None else sim_func,
        )  # type: Callable[[str, str], float]

    def sim_score(self, src: str, tar: str) -> float:
        """Return the Smith-Waterman score of two strings.

        Parameters
        ----------
        src : str
            Source string for comparison
        tar : str
            Target string for comparison

        Returns
        -------
        float
            Smith-Waterman score

        Examples
        --------
        >>> cmp = SmithWaterman()
        >>> cmp.sim_score('cat', 'hat')
        2.0
        >>> cmp.sim_score('Niall', 'Neil')
        1.0
        >>> cmp.sim_score('aluminum', 'Catalan')
        0.0
        >>> cmp.sim_score('ATCG', 'TAGC')
        1.0


        .. versionadded:: 0.1.0
        .. versionchanged:: 0.3.6
            Encapsulated in class

        """
        d_mat = np_zeros((len(src) + 1, len(tar) + 1), dtype=np_float)

        for i in range(1, len(src) + 1):
            for j in range(1, len(tar) + 1):
                match = d_mat[i - 1, j - 1] + self._sim_func(
                    src[i - 1], tar[j - 1]
                )
                delete = d_mat[i - 1, j] - self._gap_cost
                insert = d_mat[i, j - 1] - self._gap_cost
                d_mat[i, j] = max(0, match, delete, insert)
        return cast(float, d_mat[d_mat.shape[0] - 1, d_mat.shape[1] - 1])

    def sim(self, src: str, tar: str) -> float:
        """Return the normalized Smith-Waterman score of two strings.

        Parameters
        ----------
        src : str
            Source string for comparison
        tar : str
            Target string for comparison

        Returns
        -------
        float
            Normalized Smith-Waterman score

        Examples
        --------
        >>> cmp = SmithWaterman()
        >>> cmp.sim('cat', 'hat')
        0.6666666666666667
        >>> cmp.sim('Niall', 'Neil')
        0.22360679774997896
        >>> round(cmp.sim('aluminum', 'Catalan'), 12)
        0.0
        >>> cmp.sim('cat', 'hat')
        0.6666666666666667


        .. versionadded:: 0.4.1

        """
        if src == tar:
            return 1.0
        return max(0.0, self.sim_score(src, tar)) / (
            self.sim_score(src, src) ** 0.5 * self.sim_score(tar, tar) ** 0.5
        )


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# Copyright 2019-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.tests.distance.test_distance__token_distance.

This module contains unit tests for abydos.distance._TokenDistance
"""

import unittest
from collections import Counter

from abydos.distance import (
    AverageLinkage,
    DamerauLevenshtein,
    Jaccard,
    JaroWinkler,
    SokalMichener,
)
from abydos.stats import ConfusionTable
from abydos.tokenizer import (
    CharacterTokenizer,
    QSkipgrams,
    WhitespaceTokenizer,
)


class TokenDistanceTestCases(unittest.TestCase):
    """Test _TokenDistance functions.

    abydos.distance._TokenDistance
    """

    cmp_j_crisp = Jaccard(intersection_type='crisp')
    cmp_j_soft = Jaccard(intersection_type='soft')
    cmp_j_fuzzy = Jaccard(
        intersection_type='fuzzy', metric=DamerauLevenshtein(), threshold=0.4
    )
    cmp_j_linkage = Jaccard(intersection_type='linkage')

    def test_crisp_jaccard_sim(self):
        """Test abydos.distance.Jaccard.sim (crisp)."""
        # Base cases
        self.assertEqual(self.cmp_j_crisp.sim('', ''), 1.0)
        self.assertEqual(self.cmp_j_crisp.sim('a', ''), 0.0)
        self.assertEqual(self.cmp_j_crisp.sim('', 'a'), 0.0)
        self.assertEqual(self.cmp_j_crisp.sim('abc', ''), 0.0)
        self.assertEqual(self.cmp_j_crisp.sim('', 'abc'), 0.0)
        self.assertEqual(self.cmp_j_crisp.sim('abc', 'abc'), 1.0)
        self.assertEqual(self.cmp_j_crisp.sim('abcd', 'efgh'), 0.0)

        self.assertAlmostEqual(
            self.cmp_j_crisp.sim('Nigel', 'Niall'), 0.3333333333
        )
        self.assertAlmostEqual(
            self.cmp_j_crisp.sim('Niall', 'Nigel'), 0.3333333333
        )
        self.assertAlmostEqual(
            self.cmp_j_crisp.sim('Colin', 'Coiln'), 0.3333333333
        )
        self.assertAlmostEqual(
            self.cmp_j_crisp.sim('Coiln', 'Colin'), 0.3333333333
        )
        self.assertAlmostEqual(
            self.cmp_j_crisp.sim('ATCAACGAGT', 'AACGATTAG'), 0.5
        )

    def test_soft_jaccard_sim(self):
        """Test abydos.distance.Jaccard.sim (soft)."""
        # Base cases
        self.assertEqual(self.cmp_j_soft.sim('', ''), 1.0)
        self.assertEqual(self.cmp_j_soft.sim('a', ''), 0.0)
        self.assertEqual(self.cmp_j_soft.sim('', 'a'), 0.0)
        self.assertEqual(self.cmp_j_soft.sim('abc', ''), 0.0)
        self.assertEqual(self.cmp_j_soft.sim('', 'abc'), 0.0)
        self.assertEqual(self.cmp_j_soft.sim('abc', 'abc'), 1.0)
        self.assertAlmostEqual(self.cmp_j_soft.sim('abcd', 'efgh'), 0.11111111)

        self.assertAlmostEqual(self.cmp_j_soft.sim('Nigel', 'Niall'), 0.5)
        self.assertAlmostEqual(self.cmp_j_soft.sim('Niall', 'Nigel'), 0.5)
        self.assertAlmostEqual(self.cmp_j_soft.sim('Colin', 'Coiln'), 0.6)
        self.assertAlmostEqual(self.cmp_j_soft.sim('Coiln', 'Colin'), 0.6)
        self.assertAlmostEqual(
            self.cmp_j_soft.sim('ATCAACGAGT', 'AACGATTAG'), 0.68
        )

        self.assertAlmostEqual(
            Jaccard(
                intersection_type='soft', tokenizer=WhitespaceTokenizer()
            ).sim('junior system analyst', 'systems analyst'),
            0.6190476190476191,
        )
        self.assertAlmostEqual(
            Jaccard(
                intersection_type='soft', tokenizer=WhitespaceTokenizer()
            ).sim('systems analyst', 'junior system analyst'),
            0.6190476190476191,
        )

        with self.assertRaises(TypeError):
            Jaccard(
                intersection_type='soft',
                metric=JaroWinkler(),
                tokenizer=WhitespaceTokenizer(),
            ).sim('junior system analyst', 'systems analyst')

    def test_fuzzy_jaccard_sim(self):
        """Test abydos.distance.Jaccard.sim (fuzzy)."""
        # Base cases
        self.assertEqual(self.cmp_j_fuzzy.sim('', ''), 1.0)
        self.assertEqual(self.cmp_j_fuzzy.sim('a', ''), 0.0)
        self.assertEqual(self.cmp_j_fuzzy.sim('', 'a'), 0.0)
        self.assertEqual(self.cmp_j_fuzzy.sim('abc', ''), 0.0)
        self.assertEqual(self.cmp_j_fuzzy.sim('', 'abc'), 0.0)
        self.assertEqual(self.cmp_j_fuzzy.sim('abc', 'abc'), 1.0)
        self.assertAlmostEqual(
            self.cmp_j_fuzzy.sim('abcd', 'efgh'), 0.1111111111111111
        )

        self.assertAlmostEqual(self.cmp_j_fuzzy.sim('Nigel', 'Niall'), 0.5)
        self.assertAlmostEqual(self.cmp_j_fuzzy.sim('Niall', 'Nigel'), 0.5)
        self.assertAlmostEqual(self.cmp_j_fuzzy.sim('Colin', 'Coiln'), 0.6)
        self.assertAlmostEqual(self.cmp_j_fuzzy.sim('Coiln', 'Colin'), 0.6)
        self.assertAlmostEqual(
            self.cmp_j_fuzzy.sim('ATCAACGAGT', 'AACGATTAG'), 0.68
        )
        self.assertEqual(sum(self.cmp_j_fuzzy._union().values()), 11.0)

        self.assertAlmostEqual(
            Jaccard(intersection_type='fuzzy').sim('synonym', 'antonym'),
            0.3333333333333333,
        )

    def test_linkage_jaccard_sim(self):
        """Test abydos.distance.Jaccard.sim (group linkage)."""
        # Base cases
        self.assertEqual(self.cmp_j_linkage.sim('', ''), 1.0)
        self.assertEqual(self.cmp_j_linkage.sim('a', ''), 0.0)
        self.assertEqual(self.cmp_j_linkage.sim('', 'a'), 0.0)
        self.assertEqual(self.cmp_j_linkage.sim('abc', ''), 0.0)
        self.assertEqual(self.cmp_j_linkage.sim('', 'abc'), 0.0)
        self.assertEqual(self.cmp_j_linkage.sim('abc', 'abc'), 1.0)
        self.assertAlmostEqual(
            self.cmp_j_linkage.sim('abcd', 'efgh'), 0.1111111111111111
        )

        self.assertAlmostEqual(self.cmp_j_linkage.sim('Nigel', 'Niall'), 0.5)
        self.assertAlmostEqual(self.cmp_j_linkage.sim('Niall', 'Nigel'), 0.5)
        self.assertAlmostEqual(self.cmp_j_linkage.sim('Colin', 'Coiln'), 0.6)
        self.assertAlmostEqual(self.cmp_j_linkage.sim('Coiln', 'Colin'), 0.6)
        self.assertAlmostEqual(
            self.cmp_j_linkage.sim('ATCAACGAGT', 'AACGATTAG'), 0.68
        )

        self.assertAlmostEqual(
            Jaccard(
                intersection_type='linkage',
                metric=JaroWinkler(),
                threshold=0.2,
            ).sim('synonym', 'antonym'),
            0.6,
        )

    def test_token_distance(self):
        """Test abydos.distance._TokenDistance members."""
        self.assertAlmostEqual(
            Jaccard(intersection_type='soft', alphabet=24).sim(
                'ATCAACGAGT', 'AACGATTAG'
            ),
            0.68,
        )
        self.assertAlmostEqual(
            Jaccard(qval=1, alphabet='CGAT').sim('ATCAACGAGT', 'AACGATTAG'),
            0.9,
        )
        self.assertAlmostEqual(
            Jaccard(tokenizer=QSkipgrams(qval=3), alphabet='CGAT').sim(
                'ATCAACGAGT', 'AACGATTAG'
            ),
            0.6372795969773299,
        )
        self.assertAlmostEqual(
            Jaccard(alphabet=None).sim('synonym', 'antonym'),
            0.3333333333333333,
        )
        self.assertAlmostEqual(
            Jaccard(tokenizer=QSkipgrams(qval=3)).sim('synonym', 'antonym'),
            0.34146341463414637,
        )

        src_ctr = Counter({'a': 5, 'b': 2, 'c': 10})
        tar_ctr = Counter({'a': 2, 'c': 1, 'd': 3, 'e': 12})
        self.assertAlmostEqual(Jaccard().sim(src_ctr, tar_ctr), 0.09375)

        self.assertAlmostEqual(
            SokalMichener(normalizer='proportional').sim('synonym', 'antonym'),
            0.984777917351113,
        )
        self.assertAlmostEqual(
            SokalMichener(normalizer='log').sim('synonym', 'antonym'),
            1.2385752469545532,
        )
        self.assertAlmostEqual(
            SokalMichener(normalizer='exp', alphabet=0).sim(
                'synonym', 'antonym'
            ),
            3.221246147982545e18,
        )
        self.assertAlmostEqual(
            SokalMichener(normalizer='laplace').sim('synonym', 'antonym'),
            0.98856416772554,
        )
        self.assertAlmostEqual(
            SokalMichener(normalizer='inverse').sim('synonym', 'antonym'),
            197.95790155440417,
        )
        self.assertAlmostEqual(
            SokalMichener(normalizer='complement').sim('synonym', 'antonym'),
            1.0204081632653061,
        )
        self.assertAlmostEqual(
            SokalMichener(normalizer='base case').sim('synonym', 'antonym'),
            0.9897959183673469,
        )
        self.assertAlmostEqual(
            SokalMichener().sim('synonym', 'antonym'), 0.9897959183673469
        )

        sm = SokalMichener()
        sm._tokenize('synonym', 'antonym')  # noqa: SF01

        self.assertEqual(
            sm._get_tokens(),  # noqa: SF01
            (
                Counter(
                    {
                        '$s': 1,
                        'sy': 1,
                        'yn': 1,
                        'no': 1,
                        'on': 1,
                        'ny': 1,
                        'ym': 1,
                        'm#': 1,
                    }
                ),
                Counter(
                    {
                        '$a': 1,
                        'an': 1,
                        'nt': 1,
                        'to': 1,
                        'on': 1,
                        'ny': 1,
                        'ym': 1,
                        'm#': 1,
                    }
                ),
            ),
        )
        self.assertEqual(sm._src_card(), 8)  # noqa: SF01
        self.assertEqual(sm._tar_card(), 8)  # noqa: SF01
        self.assertEqual(
            sm._symmetric_difference(),  # noqa: SF01
            Counter(
                {
                    '$s': 1,
                    'sy': 1,
                    'yn': 1,
                    'no': 1,
                    '$a': 1,
                    'an': 1,
                    'nt': 1,
                    'to': 1,
                }
            ),
        )
        self.assertEqual(sm._symmetric_difference_card(), 8)  # noqa: SF01
        self.assertEqual(sm._total_complement_card(), 772)  # noqa: SF01
        self.assertEqual(sm._population_card(), 788)  # noqa: SF01
        self.assertEqual(
            sm._union(),  # noqa: SF01
            Counter(
                {
                    '$s': 1,
                    'sy': 1,
                    'yn': 1,
                    'no': 1,
                    'on': 1,
                    'ny': 1,
                    'ym': 1,
                    'm#': 1,
                    '$a': 1,
                    'an': 1,
                    'nt': 1,
                    'to': 1,
                }
            ),
        )
        self.assertEqual(sm._union_card(), 12)  # noqa: SF01
        self.assertEqual(
            sm._difference(),  # noqa: SF01
            Counter(
                {
                    '$s': 1,
                    'sy': 1,
                    'yn': 1,
                    'no': 1,
                    'on': 0,
                    'ny': 0,
                    'ym': 0,
                    'm#': 0,
                    '$a': -1,
                    'an': -1,
                    'nt': -1,
                    'to': -1,
                }
            ),
        )
        self.assertEqual(
            sm._intersection(),  # noqa: SF01
            Counter({'on': 1, 'ny': 1, 'ym': 1, 'm#': 1}),
        )
        self.assertEqual(
            sm._get_confusion_table(),  # noqa: SF01
            ConfusionTable(tp=4, tn=772, fp=4, fn=4),
        )

        sm = SokalMichener(
            alphabet=Counter({'C': 20, 'G': 20, 'A': 20, 'T': 20}), qval=1
        )
        sm._tokenize('ATCAACGAGT', 'AACGATTAG')  # noqa: SF01
        self.assertEqual(sm._total_complement_card(), 61)  # noqa: SF01

        self.assertAlmostEqual(
            self.cmp_j_linkage.sim('abandonned', 'abandoned'),
            0.9090909090909091,
        )
        self.assertAlmostEqual(
            self.cmp_j_linkage.sim('abundacies', 'abundances'),
            0.6923076923076923,
        )

        # Some additional constructors needed to complete test coverage
        self.assertAlmostEqual(
            Jaccard(alphabet=None, qval=range(2, 4)).sim('abc', 'abcd'),
            0.42857142857142855,
        )
        self.assertAlmostEqual(
            AverageLinkage(qval=range(2, 4)).sim('abc', 'abcd'),
            0.22558922558922556,
        )
        self.assertAlmostEqual(
            Jaccard(alphabet='abcdefghijklmnop', qval=range(2, 4)).sim(
                'abc', 'abcd'
            ),
            0.42857142857142855,
        )
        self.assertAlmostEqual(
            Jaccard(
                alphabet='abcdefghijklmnop', tokenizer=WhitespaceTokenizer()
            ).sim('abc', 'abcd'),
            0.0,
        )
        self.assertAlmostEqual(
            Jaccard(alphabet=list('abcdefghijklmnop')).sim('abc', 'abcd'), 0.5
        )
        self.assertAlmostEqual(
            Jaccard(tokenizer=CharacterTokenizer()).sim('abc', 'abcd'), 0.75
        )

        cmp_j_soft = Jaccard(intersection_type='soft')
        self.assertEqual(cmp_j_soft._src_card(), 0)  # noqa: SF01
        self.assertEqual(cmp_j_soft._tar_card(), 0)  # noqa: SF01
        self.assertEqual(cmp_j_soft._src_only(), Counter())  # noqa: SF01
        self.assertEqual(cmp_j_soft._tar_only(), Counter())  # noqa: SF01
        self.assertEqual(cmp_j_soft._total(), Counter())  # noqa: SF01
        self.assertEqual(cmp_j_soft._union(), Counter())  # noqa: SF01
        self.assertEqual(cmp_j_soft._difference(), Counter())  # noqa: SF01
        cmp_j_soft.sim('abcd', 'abcde')
        self.assertEqual(cmp_j_soft._src_card(), 5)  # noqa: SF01
        self.assertEqual(cmp_j_soft._tar_card(), 6)  # noqa: SF01
        self.assertEqual(
            cmp_j_soft._src_only(), Counter({'#': 0.5})  # noqa: SF01
        )
        self.assertEqual(
            cmp_j_soft._tar_only(), Counter({'e#': 1, 'e': 0.5})  # noqa: SF01
        )
        self.assertEqual(
            cmp_j_soft._total(),  # noqa: SF01
            Counter(
                {
                    'e#': 1,
                    'e': 0.5,
                    '#': 0.5,
                    '$a': 2,
                    'ab': 2,
                    'bc': 2,
                    'cd': 2,
                    'd': 1.0,
                }
            ),
        )
        self.assertEqual(
            cmp_j_soft._union(),  # noqa: SF01
            Counter(
                {
                    'e#': 1,
                    'e': 0.5,
                    '#': 0.5,
                    '$a': 1,
                    'ab': 1,
                    'bc': 1,
                    'cd': 1,
                    'd': 0.5,
                }
            ),
        )
        self.assertEqual(
            cmp_j_soft._difference(),  # noqa: SF01
            Counter({'#': 0.5, 'e#': -1, 'e': -0.5}),
        )


if __name__ == '__main__':
    unittest.main()



# Copyright 2019-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.tests.tokenizer.test_tokenizer_qgrams.

This module contains unit tests for abydos.tokenizer.QGrams
"""

import unittest

from abydos.tokenizer import NLTKTokenizer

try:
    from nltk import TweetTokenizer
except ImportError:  # pragma: no cover
    TweetTokenizer = None


class NLTKTokenizerTestCases(unittest.TestCase):
    """Test abydos.tokenizer.NLTKTokenizer."""

    def test_nltk_tokenizer(self):
        """Test abydos.tokenizer.NLTKTokenizer."""
        if TweetTokenizer is None:  # pragma: no cover
            return

        tok = NLTKTokenizer(nltk_tokenizer=TweetTokenizer())

        self.assertEqual(sorted(tok.tokenize('').get_list()), [])
        self.assertEqual(sorted(tok.tokenize('a').get_list()), ['a'])

        self.assertEqual(
            sorted(tok.tokenize('NELSON').get_list()), sorted(['NELSON'])
        )
        self.assertEqual(
            sorted(tok.tokenize('NEILSEN').get_list()), sorted(['NEILSEN'])
        )

        tok = NLTKTokenizer(nltk_tokenizer=TweetTokenizer)
        tweet1 = 'Big night of basketball - @Warriors chasing 73 and a\
        farewell for an all-timer, @KobeBryant. NBA fans feeling like:'
        self.assertEqual(
            sorted(tok.tokenize(tweet1).get_list()),
            sorted(
                [
                    'Big',
                    'night',
                    'of',
                    'basketball',
                    '-',
                    '@Warriors',
                    'chasing',
                    '73',
                    'and',
                    'a',
                    'farewell',
                    'for',
                    'an',
                    'all-timer',
                    ',',
                    '@KobeBryant',
                    '.',
                    'NBA',
                    'fans',
                    'feeling',
                    'like',
                    ':',
                ]
            ),
        )

        tweet2 = 'Einstein was right! Congrats to @NSF and @LIGO on detecting\
        gravitational waves - a huge breakthrough in how we understand the\
        universe.'
        self.assertEqual(
            sorted(tok.tokenize(tweet2).get_list()),
            sorted(
                [
                    'Einstein',
                    'was',
                    'right',
                    '!',
                    'Congrats',
                    'to',
                    '@NSF',
                    'and',
                    '@LIGO',
                    'on',
                    'detecting',
                    'gravitational',
                    'waves',
                    '-',
                    'a',
                    'huge',
                    'breakthrough',
                    'in',
                    'how',
                    'we',
                    'understand',
                    'the',
                    'universe',
                    '.',
                ]
            ),
        )

        with self.assertRaises(TypeError):
            NLTKTokenizer(nltk_tokenizer='TweetTokenizer')


if __name__ == '__main__':
    unittest.main()



# Copyright 2014-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.compression._rle.

Run-Length Encoding encoder/decoder
"""

from itertools import groupby

__all__ = ['RLE']


class RLE:
    """Run-Length Encoding.

    Cf. :cite:`Robinson:1967`.

    Based on http://rosettacode.org/wiki/Run-length_encoding#Python
    :cite:`rosettacode:2018`. This is licensed GFDL 1.2.

    Digits 0-9 cannot be in text.

    .. versionadded:: 0.3.6
    """

    def encode(self, text: str) -> str:
        r"""Perform encoding of run-length-encoding (RLE).

        Parameters
        ----------
        text : str
            A text string to encode

        Returns
        -------
        str
            Word decoded by RLE

        Examples
        --------
        >>> from abydos.compression import BWT
        >>> rle = RLE()
        >>> bwt = BWT()
        >>> rle.encode(bwt.encode('align'))
        'n\x00ilag'
        >>> rle.encode('align')
        'align'

        >>> rle.encode(bwt.encode('banana'))
        'annb\x00aa'
        >>> rle.encode('banana')
        'banana'

        >>> rle.encode(bwt.encode('aaabaabababa'))
        'ab\x00abbab5a'
        >>> rle.encode('aaabaabababa')
        '3abaabababa'

        .. versionadded:: 0.1.0
        .. versionchanged:: 0.3.6
            Encapsulated in class

        """
        if text:
            text = ''.join(
                (str(n) + k if n > 2 else (k if n == 1 else 2 * k))
                for n, k in ((len(list(g)), k) for k, g in groupby(text))
            )
        return text

    def decode(self, text: str) -> str:
        r"""Perform decoding of run-length-encoding (RLE).

        Parameters
        ----------
        text : str
            A text string to decode

        Returns
        -------
        str
            Word decoded by RLE

        Examples
        --------
        >>> from abydos.compression import BWT
        >>> rle = RLE()
        >>> bwt = BWT()
        >>> bwt.decode(rle.decode('n\x00ilag'))
        'align'
        >>> rle.decode('align')
        'align'

        >>> bwt.decode(rle.decode('annb\x00aa'))
        'banana'
        >>> rle.decode('banana')
        'banana'

        >>> bwt.decode(rle.decode('ab\x00abbab5a'))
        'aaabaabababa'
        >>> rle.decode('3abaabababa')
        'aaabaabababa'

        .. versionadded:: 0.1.0
        .. versionchanged:: 0.3.6
            Encapsulated in class

        """
        mult = ''
        decoded = []
        for letter in list(text):
            if not letter.isdigit():
                if mult:
                    decoded.append(int(mult) * letter)
                    mult = ''
                else:
                    decoded.append(letter)
            else:
                mult += letter

        text = ''.join(decoded)
        return text


if __name__ == '__main__':
    import doctest

    doctest.testmod()



# Copyright 2019-2020 by Christopher C. Little.
# This file is part of Abydos.
#
# Abydos is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Abydos is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Abydos. If not, see <http://www.gnu.org/licenses/>.

"""abydos.tests.distance.test_distance_kulczynski_i.

This module contains unit tests for abydos.distance.KulczynskiI
"""

import unittest

from abydos.distance import KulczynskiI


class KulczynskiITestCases(unittest.TestCase):
    """Test KulczynskiI functions.

    abydos.distance.KulczynskiI
    """

    cmp = KulczynskiI()

    def test_kulczynski_i_sim_score(self):
        """Test abydos.distance.KulczynskiI.sim_score."""
        # Base cases
        self.assertEqual(self.cmp.sim_score('', ''), 0.0)
        self.assertEqual(self.cmp.sim_score('a', ''), 0.0)
        self.assertEqual(self.cmp.sim_score('', 'a'), 0.0)
        self.assertEqual(self.cmp.sim_score('abc', ''), 0.0)
        self.assertEqual(self.cmp.sim_score('', 'abc'), 0.0)
        self.assertEqual(self.cmp.sim_score('abc', 'abc'), float('inf'))
        self.assertEqual(self.cmp.sim_score('abcd', 'efgh'), 0.0)

        self.assertAlmostEqual(self.cmp.sim_score('Nigel', 'Niall'), 0.5)
        self.assertAlmostEqual(self.cmp.sim_score('Niall', 'Nigel'), 0.5)
        self.assertAlmostEqual(self.cmp.sim_score('Colin', 'Coiln'), 0.5)
        self.assertAlmostEqual(self.cmp.sim_score('Coiln', 'Colin'), 0.5)
        self.assertAlmostEqual(
            self.cmp.sim_score('ATCAACGAGT', 'AACGATTAG'), 1.0
        )

    def test_kulczynski_i_dist(self):
        """Test abydos.distance.KulczynskiI.dist."""
        self.assertRaises(NotImplementedError, self.cmp.dist)

    def test_kulczynski_i_sim(self):
        """Test abydos.distance.KulczynskiI.sim."""
        self.assertRaises(NotImplementedError, self.cmp.sim)


if __name__ == '__main__':
    unittest.main()
